from abc import ABC, abstractmethod

from smartcal.config.singleton import SingletonABC


class ConfigurationManagerBase(ABC, metaclass=SingletonABC):
    """
    Configurations
    """
    @property
    @abstractmethod
    def n_bins(self):
        pass
    
    @property
    @abstractmethod
    def n_bins_calibraion_curve(self):
        pass
    
    @property
    @abstractmethod
    def random_seed(self):
        pass

    @property
    @abstractmethod
    def n_bin_cal(self):
        pass
    
    @property
    @abstractmethod
    def beta_model(self):
        pass

    @property
    @abstractmethod
    def device(self):
        pass
    
    @property
    @abstractmethod
    def patience(self):
        pass
    
    @property
    @abstractmethod
    def min_lr_img(self):
        pass
    
    @property
    @abstractmethod
    def max_lr_img(self):
        pass
    
    @property
    @abstractmethod
    def min_delta_early_stopper_img(self):
        pass
    
    @property
    @abstractmethod
    def num_itr_early_stopper_img(self):
        pass
    
    @property 
    @abstractmethod   
    def language_train_epochs(self):
        pass
    
    @property 
    @abstractmethod   
    def language_patience(self):
        pass
    
    @property   
    @abstractmethod 
    def language_min_delta(self):
        pass
    
    @property  
    @abstractmethod  
    def language_batch_size(self):
        pass
    
    @property  
    @abstractmethod  
    def language_monitor_metric(self):
        pass

    @property
    @abstractmethod
    def conf_threshold(self):
        pass
    
    @property
    @abstractmethod
    def n_bin_metrics(self):
        pass

    @property
    @abstractmethod
    def conf_threshold(self):
        pass
    
    @property
    @abstractmethod
    def conf_thresholds_list(self):
        pass
    
    @property
    @abstractmethod
    def epochs(self):
        pass
    
    @property
    @abstractmethod
    def learning_rate(self):
        pass

    @property
    @abstractmethod
    def split_ratios(self):
        pass

    @property
    @abstractmethod
    def split_ratios_images(self):
        pass

    @property
    @abstractmethod
    def batch_size(self):
        pass

    @property
    @abstractmethod
    def img_size(self):
        pass

    @property
    @abstractmethod
    def tokenizer_max_length(self):
        pass

    @property
    @abstractmethod
    def bert_tokenizer(self):
        pass

    @property
    @abstractmethod
    def config_img(self):
        pass

    @property
    @abstractmethod
    def config_tabular(self):
        pass

    @property
    @abstractmethod
    def config_language(self):
        pass
    @property
    @abstractmethod
    def lr_tempscaling(self):
        pass

    @property
    @abstractmethod
    def max_iter_tempscaling(self):
        pass

    @property
    @abstractmethod
    def lr_vectorscaling(self):
        pass

    @property
    @abstractmethod
    def max_iter_vectorscaling(self):
        pass

    @property
    @abstractmethod
    def lr_matrixscaling(self):
        pass

    @property
    @abstractmethod
    def max_iter_matrixscaling(self):
        pass

    @property
    @abstractmethod
    def initial_T(self):
        pass

    @property
    @abstractmethod
    def n_bin_meta_features(self):
        pass

    @property
    @abstractmethod
    def meta_data_file(self):
        pass

    @property    
    @abstractmethod
    def lr_dirichlet(self):
        pass

    @property    
    @abstractmethod
    def max_iter_dirichlet(self):
        pass

    @property    
    @abstractmethod
    def probability_tree_max_depth(self):
        pass

    @property    
    @abstractmethod
    def probability_tree_min_samples_leaf(self):
        pass

    @property    
    @abstractmethod
    def meta_alpha(self):
        pass

    @property    
    @abstractmethod
    def meta_acc(self):
        pass

    @property    
    @abstractmethod
    def meta_default_constraint(self):
        pass

    @property 
    @abstractmethod   
    def platt_default_type(self):
        pass
    
    @property  
    @abstractmethod  
    def histogram_default_type(self):
        pass
    
    # FOR HYPERPARAMETER TUNING
    @property    
    @abstractmethod
    def beta_params_tune(self):
        pass
    
    @property    
    @abstractmethod
    def cal_n_bins_tune(self):
        pass
    
    @property    
    @abstractmethod
    def cal_lr_tune(self):
        pass
    
    @property    
    @abstractmethod
    def cal_max_iter_tune(self):
        pass
    
    @property    
    @abstractmethod
    def meta_calibrator_calibrator_type_tune(self):
        pass

    @property    
    @abstractmethod
    def meta_alpha_tune(self):
        pass
    
    @property    
    @abstractmethod
    def meta_acc_tune(self):
        pass
    
    @property    
    @abstractmethod
    def probability_tree_max_depth_tune(self):
        pass
    
    @property    
    @abstractmethod
    def probability_tree_min_samples_leaf_tune(self):
        pass
    
    @property    
    @abstractmethod
    def temperature_scaling_initial_t_tune(self):
        pass
    
    @property    
    @abstractmethod
    def histogram_calibrator_calibrator_type_tune(self):
        pass
    
    @property    
    @abstractmethod
    def platt_calibrator_calibrator_type_tune(self):
        pass

    @property    
    @abstractmethod
    def adaptive_temp_mode_tune(self):
        pass  
    
    @property    
    @abstractmethod
    def adaptive_temp_mode_tune(self):
        pass  
    
    @property    
    @abstractmethod
    def adaptive_temp_bins_tune(self):
        pass  
    
    @property    
    @abstractmethod
    def adaptive_temp_lr_tune(self):
        pass  

    @property    
    @abstractmethod
    def adaptive_temp_max_itr_tune(self):
        pass  
        
    @property    
    @abstractmethod
    def mix_match_parametric_tune(self):
        pass  
    
    @property    
    @abstractmethod
    def mix_match_nonparametric_tune(self):
        pass  
    
    @property    
    @abstractmethod
    def cal_tune_kfolds(self):
        pass

    @property
    @abstractmethod
    def n_confidence_bins(self):
        pass

    @property
    @abstractmethod
    def n_entropy_bins(self):
        pass

    # FOR BAYESIAN OPTIMIZATION
    @property
    @abstractmethod
    def bayes_lr_tune(self):
        pass
    
    @property
    @abstractmethod
    def bayes_max_iter_tune(self):
        pass
    
    @property
    @abstractmethod
    def bayes_temp_tune(self):
        pass
    
    @property
    @abstractmethod
    def bayes_alpha_tune(self):
        pass
    
    @property
    @abstractmethod
    def bayes_acc_tune(self):
        pass
    
    @property
    @abstractmethod
    def bayes_num_bins_tune(self):
        pass
    
    @property
    @abstractmethod
    def meta_model_version(self):
        pass
    
    @property
    @abstractmethod
    def meta_model_k(self):
        pass
    
    @property
    @abstractmethod
    def num_itr_eval(self):
        pass

    @property
    @abstractmethod
    def metric(self):
        pass
    
    @property
    @abstractmethod
    def supported_metrics(self):
        pass
    
    @property
    @abstractmethod
    def k_recommendations(self):
        pass

    @property
    @abstractmethod
    def meta_label_encoder_path(self):
        pass
    
    @property
    @abstractmethod
    def logging(self):
        pass

    @property
    @abstractmethod
    def insights_folder_path(self):
        pass

    @property
    @abstractmethod
    def all_class_distributions_file(self):
        pass



from abc import ABC, abstractmethod
import numpy as np

from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager


config_manager = ConfigurationManager()

class BaseMetaModel(ABC):
	"""
    Abstract base class for meta models that predict the best calibration methods.
    Provides interface and selection logic for top-N recommendations.
    """

	def __init__(self, metric: str = config_manager.metric, top_n: int =config_manager.k_recommendations):
		"""
        Initialize the base meta model with selection criteria.

        :param metric: The calibration metric to use (e.g., 'ECE', 'MCE', etc.)
        :param top_n: Number of top models to return (default from config).
        :raises ValueError: If top_n is not in the range 1-12.
        """
		if top_n <= 0 or top_n > 12:
			raise ValueError("top_n must be a positive integer between 1 and 12")

		self.metric = metric
		self.top_n = top_n

	def _select_and_normalize(self, probabilities: np.ndarray, class_names: np.ndarray) -> list:
		"""
        Select classes based on probabilities and criteria, then normalize probabilities.

        :param probabilities: Array of probabilities for each class.
        :param class_names: Array of class names corresponding to the probabilities.
        :return: List of (class_name, normalized_probability) tuples.
        """
		sorted_indices = np.argsort(-probabilities)
		sorted_probs = probabilities[sorted_indices]
		sorted_classes = class_names[sorted_indices]

		# Apply selection criteria
		if self.top_n is not None:
			selected_probs = sorted_probs[:self.top_n]
			selected_classes = sorted_classes[:self.top_n]
		else:
			selected_probs = sorted_probs
			selected_classes = sorted_classes

		# Normalize probabilities
		total = selected_probs.sum()
		normalized_probs = selected_probs if total == 0 else selected_probs / total

		return list(zip(selected_classes, normalized_probs))

	@abstractmethod
	def predict_best_model(self, input_features: dict) -> list:
		"""
        Predict the best models with probabilities based on input features.

        :param input_features: Dictionary of input features.
        :return: List of (class_name, probability) tuples.
        """
		pass



import os
import numpy as np
import pandas as pd
from collections import Counter
from scipy.stats import entropy, skew, kurtosis
from scipy.stats import wasserstein_distance
from scipy.spatial.distance import jensenshannon

from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.utils.cal_metrics import compute_calibration_metrics
from smartcal.utils.classification_metrics import compute_classification_metrics
from smartcal.utils.labels_and_probabilities import normalize_probabilities


config_manager = ConfigurationManager()

class MetaFeaturesExtractor:
    """
    A class for processing dataset features, computing model calibration metrics,
    and performing statistical analysis.
    """

    def __init__(self):
        """
        Initializes dataset attributes, model information, calibration parameters, 
        and performance metrics.
        """

    def _process_dataset_info(self, y_true, y_pred_prob):
        y_true = np.array(y_true)  # Convert to NumPy array if not already
        y_pred_prob = np.array(y_pred_prob)  # Convert to NumPy array

        # Determine number of classes
        self.num_classes = y_pred_prob.shape[1]

        # Determine number of instances
        self.num_instances = len(y_true)
        
        if(self.num_classes == 2):
            self.dataset_type = 0 # binary
        else:
            self.dataset_type = 1 # multi-class

    def _imbalance_ratio(self, labels):
        """
        Computes the imbalance ratio of a dataset.

        Parameters:
        - labels (list): List of labels in a classification task.

        Returns:
        - tuple: (overall imbalance ratio, per-class imbalance ratios)
        """
        label_counts = Counter(labels)
        max_count = max(label_counts.values())
        min_count = min(label_counts.values())

        overall_imbalance = round(max_count / min_count, 2) if min_count > 0 else float("inf")
        class_ratios = {label: round(max_count / count, 2) for label, count in label_counts.items()}

        return overall_imbalance, class_ratios
    
    def _get_relative_frequencies(self, data):
        """
        Computes the relative frequencies of a dataset.

        Parameters:
        - data (list or array): Input data.

        Returns:
        - np.array: Relative frequency distribution.
        """
        hist, _ = np.histogram(data, bins=config_manager.n_bin_meta_features, density=False)
        return hist / hist.sum() if hist.sum() != 0 else hist

    def _calculate_statistics(self, data, prefix):
        """
        Computes and stores statistical measures for a given dataset.

        Parameters:
        - data (array-like): Input numerical data.
        - prefix (str): Prefix for attribute names where statistics will be stored.

        Returns:
        - None (Updates instance attributes dynamically).
        """
        try:
            data = np.array(data)
            if len(data) == 0:
                return

            setattr(self, f"{prefix}_Mean", round(np.mean(data), 5))
            setattr(self, f"{prefix}_Median", round(np.median(data), 5))
            setattr(self, f"{prefix}_Std", round(np.std(data), 5))
            setattr(self, f"{prefix}_Var", round(np.var(data), 5))
            setattr(self, f"{prefix}_Entropy", round(entropy(data / np.sum(data), base=2) if np.sum(data) != 0 else 0, 5))

            # If there's only one element or variance is zero, return -999 (undefined)
            if len(data) <= 1 or round(np.var(data), 5) == 0:
                setattr(self, f"{prefix}_Skewness", -999)
                setattr(self, f"{prefix}_Kurtosis", -999)
            else:
                setattr(self, f"{prefix}_Skewness", round(skew(data), 5))
                setattr(self, f"{prefix}_Kurtosis", round(kurtosis(data), 5))

            setattr(self, f"{prefix}_Min", round(np.min(data), 5))
            setattr(self, f"{prefix}_Max", round(np.max(data), 5))
        except Exception as e:
            print(f"Error computing statistics for {prefix}: {e}")

    def _calculate_confidence_differences(self, y_true, y_pred_prob):
        """
        Computes confidence values and their differences.

        Parameters:
        - y_true (list): True labels.
        - y_pred_prob (list): Predicted probabilities.

        Returns:
        - tuple: (confidence values, absolute differences between consecutive confidences)
        """
        confidence_values = [1 - y_pred_prob[i][y_true[i]] for i in range(len(y_true))]
        differences = np.abs(np.diff(confidence_values))
        return confidence_values, differences
    
    def _get_actual_predictions(self, y_pred_prob):
        """
        Returns the predicted class labels based on the highest probability.

        Parameters:
        - y_pred_prob (list): List of predicted probability distributions.

        Returns:
        - list: Predicted class labels.
        """
        return np.argmax(y_pred_prob, axis=1)  # Returns the index of the max probability for each sample
    
    def _calculate_classification_errors(self, y_true, y_pred, y_pred_prob):
        """Computes and stores classification performance metrics including micro, macro, and weighted scores."""
        try:
            metrics = compute_classification_metrics( y_true, y_pred, y_pred_prob, zero_division=0)

            # Log loss
            setattr(self, "Classification_Log_loss", round(metrics["loss"], 5))

            # Accuracy
            setattr(self, "Classification_Accuracy", round(metrics["accuracy"], 5))

            # Precision
            setattr(self, "Classification_Precision_Micro", round(metrics["precision_micro"], 5))
            setattr(self, "Classification_Precision_Macro", round(metrics["precision_macro"], 5))
            setattr(self, "Classification_Precision_Weighted", round(metrics["precision_weighted"], 5))

            # Recall
            setattr(self, "Classification_Recall_Micro", round(metrics["recall_micro"], 5))
            setattr(self, "Classification_Recall_Macro", round(metrics["recall_macro"], 5))
            setattr(self, "Classification_Recall_Weighted", round(metrics["recall_weighted"], 5))

            # F1 Score
            setattr(self, "Classification_F1_Micro", round(metrics["f1_micro"], 5))
            setattr(self, "Classification_F1_Macro", round(metrics["f1_macro"], 5))
            setattr(self, "Classification_F1_Weighted", round(metrics["f1_weighted"], 5))

        except Exception as e:
            print(f"Error computing classification metrics: {e}")

    def _calculate_calibration_metrics(self, y_pred_prob, actual_predictions, y_true):
        metrics = compute_calibration_metrics(y_pred_prob, actual_predictions, y_true, metrics_to_compute = ['ece', 'mce', 'conf_ece', 'brier_score'])

        self.ECE_before = round(metrics["ece"], 5)
        self.MCE_before = round(metrics["mce"], 5)
        self.ConfECE_before = round(metrics["conf_ece"][0], 5)
        self.brier_score_before = round(metrics["brier_score"], 5)

    def _bhattacharyya_distance(self, p, q):
        """
        Computes the Bhattacharyya distance between two probability distributions.

        Parameters:
        - p (array-like): First probability distribution.
        - q (array-like): Second probability distribution.

        Returns:
        - float: Bhattacharyya distance.
        """
        return -np.log(np.sum(np.sqrt(p * q)))
    
    def _calculate_pairwise_distances(self, data):
        """
        Computes various distance metrics between pairs of probability distributions.
        """
        num_lists = len(data)
        epsilon = 1e-10  # Small constant to prevent division errors
        distances = {"Wasserstein": [], "KL_Divergence": [], "Jensen_Shannon": [], "Bhattacharyya": []}

        try:
            for i in range(num_lists):
                for j in range(i + 1, num_lists):
                    l1 = np.array(data[i]) / (np.sum(data[i]) + epsilon)
                    l2 = np.array(data[j]) / (np.sum(data[j]) + epsilon)

                    distances["Wasserstein"].append(round(wasserstein_distance(l1, l2), 5))
                    distances["KL_Divergence"].append(round(entropy(l1, l2 + epsilon), 5))
                    distances["Jensen_Shannon"].append(round(jensenshannon(l1, l2), 5))
                    distances["Bhattacharyya"].append(round(-np.log(np.sum(np.sqrt(l1 * l2)) + epsilon), 5))
        except Exception as e:
            print(f"Error computing pairwise distances: {e}")

        return distances

    def process_features(self, y_true, y_pred_prob):
        """
        Processes dataset features, computes statistical metrics, and model calibration details.

        Returns:
        - dict: Extracted meta-features.
        """
        # Process dataset info
        self._process_dataset_info(y_true, y_pred_prob)

        # Normalize y_pred_prob if needed
        y_pred_prob = normalize_probabilities(y_pred_prob)

        # Compute imbalance ratios
        overall_imb, class_imb = self._imbalance_ratio(y_true)
        self.class_imbalance_ratio = overall_imb

        # Compute actual predictions statistics
        actual_predictions = self._get_actual_predictions(y_pred_prob)
        actual_relative_frequencies = self._get_relative_frequencies(actual_predictions)
        self.actual_predictions_entropy = entropy(actual_relative_frequencies / np.sum(actual_relative_frequencies), base=2) if np.sum(actual_relative_frequencies) != 0 else 0
        self.actual_predictions_entropy = round(self.actual_predictions_entropy, 5)

        # Compute confidence-based statistics
        confidence_values, confidence_differences = self._calculate_confidence_differences(y_true, y_pred_prob)
        self._calculate_statistics(confidence_values, "Confidence")

        # Compute Classification errors
        self._calculate_classification_errors(y_true, actual_predictions, y_pred_prob)

        # Compute calibration metrics (before)
        self._calculate_calibration_metrics(y_pred_prob, actual_predictions, y_true)

        # Compute pairwise distances between relative frequencies
        y_pred_classes = [list(row) for row in zip(*y_pred_prob)]
        relative_frequencies = [self._get_relative_frequencies(class_) for class_ in y_pred_classes]
        pairwise_distances = self._calculate_pairwise_distances(relative_frequencies)

        for metric_name, distance_list in pairwise_distances.items():
            self._calculate_statistics(distance_list, metric_name)

        # Convert stored attributes into a dictionary
        meta_features_dict = {key: value for key, value in vars(self).items() if not callable(value) and not key.startswith("_")}

        return meta_features_dict

    def set_dataset_name(self, dataset_name):
        self.Dataset_name = dataset_name

    def set_model_name(self, model_name):
        self.Model_name = model_name

    def set_calibration_metric(self, calibration_metric):
        self.Calibration_metric = calibration_metric
    
    def set_best_cal(self, best_cal):
        self.Best_Cal = best_cal

    def save_to_csv(self, csv_filename=config_manager.meta_data_file):
        """
        Saves the processed feature data to a CSV file.

        Parameters:
        - csv_filename (str): Path to the CSV file.
        """
        df = pd.DataFrame([vars(self)])

        try:
            if os.path.exists(csv_filename) and os.path.getsize(csv_filename) > 0:
                existing_df = pd.read_csv(csv_filename)
                df = pd.concat([existing_df, df], ignore_index=True)
        except pd.errors.EmptyDataError:
            print(f"Warning: '{csv_filename}' is empty. Overwriting with new data.")

        df.to_csv(csv_filename, index=False)
        #print(f"CSV file '{csv_filename}' updated successfully!")


import numpy as np
from betacal import BetaCalibration as BetaCal

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.config.enums.calibration_error_code_enum import CalibrationErrorCode
from smartcal.utils.timer import time_operation


config = ConfigurationManager()

class BetaCalibrator(CalibratorBase):
    """
    Wrapper that uses the 'betacal' library for Beta Calibration,
    handling both binary and multi-class (one-vs-rest).

    model_type
    ----------
    model_type : str (default="abm")
        Which Beta model from betacal to use: "abm", "am", or "ab".

    Attributes
    ----------
    calibrators_ : dict
        For binary classification, a single entry calibrators_[1].
        For multi-class, calibrators_[class_idx] -> BetaCalibration instance.
    n_classes_ : int
        Number of classes seen during fit.
    fitted : bool
        Whether the calibrator has been fit.
    """

    def __init__(self, model_type=config.beta_model, seed=CalibratorBase.DEFAULT_SEED):
        super().__init__(seed=seed)
        
        self.model_type = model_type
        self.calibrators_ = {}
        self.n_classes_ = None
        self.timing = {} 
        self.metadata.update({
            "calibration_type": "beta"
        })
        self.metadata["params"].update({
            "beta_model_type": model_type
        })

    @time_operation
    def fit(self, probabilities, labels):
        """
        Fit Beta calibration. If n_classes=2 => calibrate only the 'positive' column.
        Otherwise do one-vs-rest for multi-class, then re-normalize at predict time.

        model_type
        ----------
        probabilities : np.ndarray, shape (n_samples, n_classes)
            Predicted probabilities for each class.
        labels : array-like, shape (n_samples,)
            True labels (0..n_classes-1).
        """
        probabilities, labels = super().validate_inputs(probabilities, labels)

        n_samples, n_classes = probabilities.shape
        if len(labels) != n_samples:
            raise ValueError("Labels length must match number of probability rows.")

        self.n_classes_ = n_classes
        self.metadata["dataset_info"].update({
            "n_samples": n_samples,
            "n_classes": n_classes,
        })
        
        self.logger.info(
            f"BetaCalibrator.fit started with {n_samples} samples, {n_classes} classes, mode={self.model_type}")

        # --- Binary case: only calibrate column=1 ---
        if n_classes == 2:
            p_pos = probabilities[:, 1]
            y_bin = (labels == 1).astype(int)  # binary 0/1
            cal = BetaCal(parameters=self.model_type)
            cal.fit(p_pos, y_bin)
            self.calibrators_[1] = cal

        else:
            # --- Multi-class: one-vs-rest for each class ---
            for c in range(n_classes):
                p_c = probabilities[:, c]
                y_c = (labels == c).astype(int)
                cal = BetaCal(parameters=self.model_type)
                cal.fit(p_c, y_c)
                self.calibrators_[c] = cal

        self.fitted = True
        self.logger.info(f"Fitted BetaCalibrator with {len(self.calibrators_)} calibrators")

        return self

    @time_operation
    def predict(self, probabilities):
        """
        Predict calibrated probabilities.

        model_type
        ----------
        probabilities : np.ndarray, shape (n_samples, n_classes)
            Uncalibrated probabilities to calibrate.

        Returns
        -------
        np.ndarray, shape (n_samples, n_classes)
            The calibrated probabilities, row-wise summing to 1.
        """
        if not self.fitted:
            raise RuntimeError("BetaCalibrator not fitted. Call `fit` first.")

        probabilities = np.asarray(probabilities)
        if probabilities.shape[1] != self.n_classes_:
            raise ValueError("Number of columns in `probabilities` differs from training.")

        n_samples, n_classes = probabilities.shape
        out = np.zeros_like(probabilities, dtype=float)

        # Raise error if probabilities are almost constant
        def is_almost_constant(arr, tol=1e-3):
            return np.std(arr) < tol
        
        if n_classes == 2:
            cal = self.calibrators_[1]

            try:
                p_pos = cal.predict(probabilities[:, 1])  # shape (n_samples,)
            except ValueError as e:
                raise ValueError(
                    (CalibrationErrorCode.ALMOST_CONSTANT_PROBS,
                    f"BetaCalibrator (mode='{self.model_type}') failed with this data. "
                    "Try a different mode or adjust your model to produce more varied probabilities.")
                ) from e

            out[:, 1] = p_pos
            out[:, 0] = 1.0 - p_pos
            return out

        # Multi-class => one-vs-rest + re-normalize
        for c in range(n_classes):
            cal = self.calibrators_[c]

            try:
                p_c = probabilities[:, c]
                out[:, c] = cal.predict(p_c)
            except ValueError as e:
                raise ValueError(
                    (CalibrationErrorCode.ALMOST_CONSTANT_PROBS,
                    f"BetaCalibrator (mode='{self.model_type}') failed on class {c}. "
                    "Try a different mode or adjust your model to produce more varied probabilities.")
                ) from e

        # Re-normalize so each row sums to 1
        row_sums = out.sum(axis=1, keepdims=True)
        row_sums = np.clip(row_sums, 1e-15, None)
        out /= row_sums

        return out
        
    def get_timing(self):
        """Return the timing information for fit and predict operations."""
        return self.timing


from enum import Enum

from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager


config_manager = ConfigurationManager()

class CalibrationHyperparameters(Enum):
    lr = config_manager.cal_lr_tune
    max_itr = config_manager.cal_max_iter_tune
    n_bins = config_manager.cal_n_bins_tune
    num_bins = config_manager.cal_n_bins_tune
    model_type = config_manager.beta_params_tune # for beta will modify the name
    calibrator_type_meta = config_manager.meta_calibrator_calibrator_type_tune
    alpha = config_manager.meta_alpha_tune
    acc = config_manager.meta_acc_tune
    initial_T = config_manager.temperature_scaling_initial_t_tune
    lr_tempscaling = config_manager.cal_lr_tune
    max_iter_tempscaling = config_manager.cal_max_iter_tune
    max_depth = config_manager.probability_tree_max_depth_tune
    min_samples_leaf = config_manager.probability_tree_min_samples_leaf_tune
    calibrator_type_platt = config_manager.platt_calibrator_calibrator_type_tune
    calibrator_type_histogram = config_manager.histogram_calibrator_calibrator_type_tune
    adapt_temp_scaling_mode = config_manager.adaptive_temp_mode_tune
    adapt_temp_scaling_lr = config_manager.adaptive_temp_lr_tune
    adapt_temp_scaling_max_itr = config_manager.adaptive_temp_max_itr_tune
    adapt_temp_scaling_bins = config_manager.adaptive_temp_bins_tune
    mix_match_parametric = config_manager.mix_match_parametric_tune
    mix_match_nonparametric = config_manager.mix_match_nonparametric_tune


    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"



import numpy as np
from scipy.special import softmax
from sklearn.tree import DecisionTreeClassifier
from sklearn.linear_model import LogisticRegression

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.utils.timer import time_operation


config_manager = ConfigurationManager()

class ProbabilityTreeCalibrator(CalibratorBase):
    """
    Implements Probability Calibration Trees (PCT) for Multiclass Calibration.
    """
    def __init__(self, max_depth=config_manager.probability_tree_max_depth, min_samples_leaf=config_manager.probability_tree_min_samples_leaf, seed=CalibratorBase.DEFAULT_SEED):
        super().__init__(seed)
        self.tree = DecisionTreeClassifier(max_depth=max_depth, min_samples_leaf=min_samples_leaf)
        self.logit_models = {}  # Logistic regression models at leaf nodes
        self.global_logit_model = None  # Global fallback model
        self.metadata["params"].update({
            "max_depth": max_depth,
            "min_samples_leaf": min_samples_leaf,
        })
        self.timing = {}
    
    def log_odds_transform(self, probs):
        """Applies log-odds transformation to probability estimates."""
        probs = np.clip(probs, 1e-6, 1 - 1e-6)  # Avoid division by zero
        return np.log(probs / (1 - probs))
    
    @time_operation
    def fit(self, X, logits, y):
        """
        Trains the probability calibration tree.
        
        Args:
            X (np.ndarray): Feature matrix (n_samples, n_features)
            logits (np.ndarray): Uncalibrated probability estimates (n_samples, n_classes)
            y (np.ndarray): True labels (n_samples,)
        """
        X, y = self.validate_inputs(X, y)
        logits = np.asarray(logits)
        
        # Fit the decision tree on original features
        self.tree.fit(X, y)
        
        # Get leaf nodes for each training instance
        leaf_indices = self.tree.apply(X)
        
        # Train a global fallback logistic regression model on all data
        global_X_log_odds = self.log_odds_transform(softmax(logits, axis=1))
        self.global_logit_model = LogisticRegression(solver='lbfgs', max_iter=1000)
        self.global_logit_model.fit(global_X_log_odds, y)
        
        # Train logistic regression models at each leaf node
        for leaf in np.unique(leaf_indices):
            indices = np.where(leaf_indices == leaf)[0]
            y_leaf = y[indices]
            
            # Check if we have enough samples AND class diversity
            if len(indices) < 2 or len(np.unique(y_leaf)) < 2:
                # Use global model if not enough samples or only one class in leaf
                self.logit_models[leaf] = self.global_logit_model
                continue
            
            X_leaf = logits[indices]  # Use logits as features for logistic regression
            
            # Convert logits to log-odds
            X_leaf_log_odds = self.log_odds_transform(softmax(X_leaf, axis=1))
            
            # Fit logistic regression model for calibration
            model = LogisticRegression(solver='lbfgs', max_iter=1000)
            model.fit(X_leaf_log_odds, y_leaf)
            self.logit_models[leaf] = model
        
        self.metadata["dataset_info"].update({
            "n_samples": X.shape[0],
            "n_features": X.shape[1]
        })
        self.fitted = True
    
    @time_operation
    def predict(self, X, logits):
        """
        Calibrates probability estimates using the learned tree and logistic models.
        
        Args:
            X (np.ndarray): Feature matrix (n_samples, n_features)
            logits (np.ndarray): Uncalibrated probability estimates (n_samples, n_classes)
        
        Returns:
            np.ndarray: Calibrated probability estimates (n_samples, n_classes)
        """
        if not self.fitted:
            raise RuntimeError("Model must be fitted before prediction.")
        
        X = np.asarray(X)
        logits = np.asarray(logits)
        n_classes = logits.shape[1]
        
        # Get leaf nodes for each test instance
        leaf_indices = self.tree.apply(X)
        calibrated_probs = np.zeros_like(logits)
        
        for i, leaf in enumerate(leaf_indices):
            model = self.logit_models.get(leaf, self.global_logit_model)  # Use leaf model or fallback
            log_odds = self.log_odds_transform(softmax(logits[i].reshape(1, -1), axis=1))
            pred_probs = model.predict_proba(log_odds)
            
            # Ensure shape compatibility with expected number of classes
            if pred_probs.shape[1] < n_classes:
                missing_classes = n_classes - pred_probs.shape[1]
                pred_probs = np.hstack([pred_probs, np.full((1, missing_classes), 1e-6)])
            
            calibrated_probs[i] = softmax(pred_probs)
        
        return calibrated_probs
    
    def get_timing(self):
        return self.timing


from enum import Enum


class ConfigReaderEnum(Enum):
    env = "EnvConfigReader"

    @staticmethod
    def get_value(name):
        for choice in ConfigReaderEnum:
            if choice.name == name:
                return choice.value
        raise ValueError(f"{name} is not a valid ConfigReaderEnum")



import numpy as np
from typing import Optional

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.utils.timer import time_operation


config_manager = ConfigurationManager()

class EmpiricalBinningCalibrator(CalibratorBase):
    def __init__(self, n_bins: int = config_manager.n_bin_cal):
        super().__init__()
        self.validate_init_params(n_bins)

        self.n_bins = n_bins
        self.bin_boundaries = None
        self.bin_probabilities = None
        self.timing = {}

        self.logger.info(f"Initialized EmpiricalBinningCalibrator with {n_bins} bins")

        # Update metadata with specific parameters
        self.metadata.update({
            'calibration_type': 'empirical_binning'
        })
        self.metadata["params"].update({
            'n_bins': n_bins
        })

    def validate_init_params(self, n_bins: int) -> None:
        if not isinstance(n_bins, int):
            raise TypeError("n_bins must be an integer")
        if n_bins <= 0:
            raise ValueError("n_bins must be positive")

    def validate_input(self, predictions, ground_truth=None):
        """
        Extends the base class validation with multiclass classification checks.
        """
        # First use the base class validation
        if ground_truth is not None:
            predictions, ground_truth = super().validate_inputs(predictions, ground_truth)

        # Ensure predictions are properly shaped
        if predictions.ndim != 2:
            raise ValueError("Predictions should be a 2D array (n_samples, n_classes)")

        n_classes = predictions.shape[1]
        if np.any((predictions < 0) | (predictions > 1)):
            raise ValueError("Predictions must be in range [0, 1]")

        if ground_truth is not None:
            unique_labels = np.unique(ground_truth)
            if not np.all(np.isin(unique_labels, np.arange(n_classes))):
                raise ValueError(f"Ground truth labels must be integers from 0 to {n_classes - 1}")

        return predictions, ground_truth


    def compute_bin_statistics(self, predictions, ground_truth: Optional[np.ndarray] = None):
        try:
            """
            Compute bin statistics including boundaries and mean probabilities.
            """

            n_classes = predictions.shape[1]
            bin_boundaries = np.linspace(0, 1, self.n_bins + 1)
            bin_probabilities = np.zeros((self.n_bins, n_classes))
            bin_counts = np.zeros(self.n_bins)  # Track number of samples in each bin

            for class_idx in range(n_classes):
                class_preds = predictions[:, class_idx]
                bin_indices = np.digitize(class_preds, bin_boundaries, right=False) - 1
                bin_indices = np.clip(bin_indices, 0, self.n_bins - 1)  # Fix indexing issue

                for i in range(self.n_bins):
                    bin_mask = bin_indices == i
                    bin_counts[i] += np.sum(bin_mask)

                    if np.any(bin_mask):
                        # Use empirical probability (fraction of true positives)
                        bin_probabilities[i, class_idx] = np.mean(ground_truth[bin_mask] == class_idx)
                    else:
                        # Mark empty bins for interpolation
                        bin_probabilities[i, class_idx] = np.nan

            # Handle empty bins through interpolation
            for class_idx in range(n_classes):
                empty_bins = np.isnan(bin_probabilities[:, class_idx])
                valid_bins = ~empty_bins

                if np.all(empty_bins):
                    bin_probabilities[:, class_idx] = 0  # Default fallback

                elif np.any(empty_bins) and np.any(valid_bins):
                    # Use linear interpolation
                    x = np.where(valid_bins)[0]
                    y = bin_probabilities[valid_bins, class_idx]
                    empty_x = np.where(empty_bins)[0]

                    bin_probabilities[empty_bins, class_idx] = np.interp(empty_x, x, y)

                # Enforce monotonicity
                bin_probabilities[:, class_idx] = np.maximum.accumulate(bin_probabilities[:, class_idx])

            # Log statistics
            self.logger.info(f"Bin sample counts: {bin_counts}")
            self.logger.info(f"Empty bins: {np.sum(bin_counts == 0)}")

            return bin_boundaries, bin_probabilities

        except Exception as e:
            self.logger.error(f"Error in compute_bin_statistics: {str(e)}")
            raise

    @time_operation
    def fit(self, predictions: np.ndarray, ground_truth):
        """
        Fit the calibrator using uncalibrated predictions and true labels.
        """
        try:
            predictions, ground_truth = self.validate_input(predictions, ground_truth)
            n_classes = predictions.shape[1]

            # Log dataset details
            self.metadata["dataset_info"].update({
            "n_samples": predictions.shape[0],
            "n_classes": predictions.shape[1],
            })

            self.logger.info(f"Starting calibration fitting with parameters: {self.metadata}")

            self.bin_boundaries, self.bin_probabilities = self.compute_bin_statistics(predictions, ground_truth)
            self.fitted = True

            # Log calibration parameters
            self.metadata['calibration_params'] = {
                'bin_boundaries': self.bin_boundaries.tolist(),
                'bin_probabilities': self.bin_probabilities.tolist()
            }

            self.logger.info(f"Calibration parameters: {self.metadata['calibration_params']}")
            self.logger.info("Successfully fitted EmpiricalBinningCalibrator")
            return self

        except Exception as e:
            self.logger.error(f"Error during fitting: {str(e)}")
            raise

    @time_operation
    def predict(self, predictions):
        """
        Calibrate predictions using the fitted parameters while preserving monotonicity.
        """   
        if not self.fitted:
            raise ValueError("Calibrator must be fitted before making predictions")

        try:
            predictions, _ = self.validate_input(predictions)

            calibrated_probs = np.zeros_like(predictions)

            for class_idx in range(predictions.shape[1]):
                class_preds = predictions[:, class_idx]
                class_bin_indices = np.digitize(class_preds, self.bin_boundaries, right=False) - 1
                class_bin_indices = np.clip(class_bin_indices, 0, self.n_bins - 1)

                calibrated_probs[:, class_idx] = self.bin_probabilities[class_bin_indices, class_idx]

            # Normalize so that probabilities sum to 1
            row_sums = np.sum(calibrated_probs, axis=1, keepdims=True)
            valid_rows = row_sums > 0
            calibrated_probs[valid_rows[:, 0], :] /= row_sums[valid_rows[:, 0], :]

            # Handle invalid (zero) rows: fallback to uniform
            invalid_rows = ~valid_rows[:, 0]
            if np.any(invalid_rows):
                calibrated_probs[invalid_rows, :] = 1.0 / predictions.shape[1]

            if not np.all((calibrated_probs >= 0) & (calibrated_probs <= 1)):
                raise ValueError("Calibrated probabilities out of range")

            if not np.allclose(np.sum(calibrated_probs, axis=1), 1, atol=1e-4):
                raise ValueError("Probabilities do not sum to 1 after normalization")

            return calibrated_probs

        except Exception as e:
            self.logger.error(f"Error during prediction: {str(e)}")
            raise
    
    def get_timing(self):
        """Return the timing information for fit and predict operations."""
        return self.timing



from smartcal.metrics import ECE, MCE, ConfECE, brier_score, log_loss

from enum import Enum


class CalibrationMetricsEnum(Enum):
    """Enum to map metric names to their corresponding classes."""
    ECE = ECE
    MCE = MCE
    ConfECE = ConfECE
    brier_score = brier_score
    log_loss = log_loss

    @classmethod
    def get_metric_class(cls, metric_name: str):
        """Retrieve the metric class by name."""
        try:
            return cls[metric_name].value
        except KeyError:
            raise ValueError(f"Metric '{metric_name}' is not supported.")



import numpy as np
import logging

from smartcal.metrics.base import BaseCalibrationMetric
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager


config_manager = ConfigurationManager()
logging.basicConfig(level=logging.INFO, force=config_manager.logging)
logger = logging.getLogger(__name__)

class ECE(BaseCalibrationMetric):
    """Class for computing the Expected Calibration Error (ECE).

    This metric measures the difference between the predicted probabilities and the actual accuracy.

    Attributes:
        num_bins (int): The number of bins to use for calibration metrics.
        metric_name (str): The name of the metric.
    """

    def __init__(self, num_bins: int):
        """Initialize the ECE.

        Args:
            num_bins (int): The number of bins to use for calibration metrics.
        """
        super().__init__(num_bins)
        self.metric_name = "ECE"

    def compute(self, predicted_probabilities, predicted_labels, true_labels) -> float:
        """Compute the Expected Calibration Error (ECE).

        Args:
            predicted_probabilities: The predicted probabilities.
            predicted_labels: The predicted labels.
            true_labels: The true labels.

        Returns:
            float: The computed ECE value.

        Raises:
            ValueError: If the lengths of predicted_probabilities and true_labels do not match.
        """
        predicted_probabilities = self._convert_to_numpy(predicted_probabilities)
        predicted_labels = self._convert_to_numpy(predicted_labels)
        true_labels = self._convert_to_numpy(true_labels)

        if not self.validate(predicted_probabilities, true_labels):
            raise ValueError("Input lengths mismatch")

        probs = predicted_probabilities
        preds = predicted_labels
        trues = true_labels

        if probs.ndim == 1:
            probs = np.vstack([1 - probs, probs]).T

        n_samples, n_classes = probs.shape
        confidences = probs[np.arange(n_samples), preds]
        accuracies = (preds == trues).astype(float)

        bins = np.linspace(0, 1, self.num_bins + 1)
        bin_indices = np.clip(np.digitize(confidences, bins[:-1], right=False) - 1, 0, self.num_bins - 1)

        ece = 0.0
        for b in range(self.num_bins):
            mask = bin_indices == b
            if np.sum(mask) == 0:
                continue
            avg_conf = np.mean(confidences[mask])
            avg_acc = np.mean(accuracies[mask])
            ece += np.abs(avg_acc - avg_conf) * np.sum(mask)

        self.metric_value = ece / n_samples
        return self.metric_value

    def logger(self):
        """Log the metric name, value, and parameters."""
        log_data = {
            "metric_name": self.metric_name,
            "metric_value": self.metric_value,
            "parameters": {
                "num_bins": self.num_bins
            }
        }
        logger.info(log_data)
        return log_data



import numpy as np

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.utils.timer import time_operation
from smartcal.calibration_algorithms.histogram.histogram import HistogramCalibrator
from smartcal.calibration_algorithms.platt.platt import PlattCalibrator
from smartcal.calibration_algorithms.temperature_scaling import TemperatureScalingCalibrator
from smartcal.calibration_algorithms.adaptive_temperature_scaling import AdaptiveTemperatureScalingCalibrator
from smartcal.calibration_algorithms.isotonic import IsotonicCalibrator
from smartcal.calibration_algorithms.meta import MetaCalibrator
from smartcal.calibration_algorithms.beta import BetaCalibrator
from smartcal.calibration_algorithms.dirichlet import DirichletCalibrator
from smartcal.calibration_algorithms.matrix_scaling import MatrixScalingCalibrator
from smartcal.calibration_algorithms.vector_scaling import VectorScalingCalibrator
from smartcal.calibration_algorithms.empirical_binning import EmpiricalBinningCalibrator


class MixAndMatchCalibrator(CalibratorBase):
    """
    Combines parametric and non-parametric calibration methods.
    First applies a parametric calibration
    then refines with a non-parametric calibration
    -------
    parametric_calibrator :
        Must be one of the following parametric calibrators:
        - TemperatureScaling (default)
        - PlattCalibrator
        - VectorScalingCalibrator
        - MatrixScaling
        - BetaCalibrator
        - Dirichlet

    nonparametric_calibrator :
        Must be one of the following non-parametric calibrators:
        - IsotonicCalibrator (default)
        - EmpiricalBinningCalibrator
        - HistogramCalibrator

    """

    def __init__(self, parametric_calibrator = "TemperatureScalingCalibrator",nonparametric_calibrator =  "IsotonicCalibrator", seed=CalibratorBase.DEFAULT_SEED):
        super().__init__(seed)
        self.timing = {}


        parametric_mapping = {
            "TemperatureScalingCalibrator": TemperatureScalingCalibrator,
            "PlattCalibrator": PlattCalibrator,
            "VectorScalingCalibrator": VectorScalingCalibrator,
            "MatrixScalingCalibrator": MatrixScalingCalibrator,
            "BetaCalibrator": BetaCalibrator,
            "MetaCalibrator": MetaCalibrator,
            "DirichletCalibrator": DirichletCalibrator,
            "AdaptiveTemperatureScalingCalibrator": AdaptiveTemperatureScalingCalibrator
        }

        nonparametric_mapping = {
            "IsotonicCalibrator": IsotonicCalibrator,
            "EmpiricalBinningCalibrator": EmpiricalBinningCalibrator,
            "HistogramCalibrator": HistogramCalibrator
        }


        if not parametric_calibrator or not nonparametric_calibrator:
            raise ValueError("Exactly one parametric and one non-parametric calibrator must be provided.")

        self.parametric_calibrator = parametric_mapping[parametric_calibrator]()
        self.nonparametric_calibrator = nonparametric_mapping[nonparametric_calibrator]()
        self.fitted = False

        self.metadata["params"].update({
            "parametric_calibrator_type": self.parametric_calibrator.__class__.__name__,
            "nonparametric_calibrator_type": self.nonparametric_calibrator.__class__.__name__,
        })

    @time_operation
    def fit(self, predictions: np.ndarray, ground_truth: np.ndarray):
        """
        Fits the parametric calibrator first, then the non-parametric calibrator.

        Args:
            predictions (np.ndarray): Model logits or probabilities of shape (n_samples, n_classes).
            ground_truth (np.ndarray): True labels of shape (n_samples,).
        """
        self.set_random_seed(self.seed)

        predictions, ground_truth = self.validate_inputs(predictions, ground_truth)

        # Fit the parametric calibrator
        self.parametric_calibrator.fit(predictions, ground_truth)
        parametric_calibrated = self.parametric_calibrator.predict(predictions)

        # Fit the non-parametric calibrator on the parametric calibrated probabilities
        self.nonparametric_calibrator.fit(parametric_calibrated, ground_truth)
        self.fitted = True

        self.metadata["dataset_info"].update({
            "n_samples": predictions.shape[0],
            "n_classes": predictions.shape[1],
        })

    @time_operation
    def predict(self, test_data: np.ndarray) -> np.ndarray:
        """
        Generates calibrated probabilities using the mix-and-match approach.

        Args:
            test_data (np.ndarray): Uncalibrated model outputs of shape (n_samples, n_classes).

        Returns:
            np.ndarray: Calibrated probabilities of shape (n_samples, n_classes).
        """
        if not self.fitted:
            raise RuntimeError("Model must be fitted before prediction.")

        # Apply parametric calibration
        parametric_calibrated = self.parametric_calibrator.predict(test_data)

        # Apply non-parametric calibration
        return self.nonparametric_calibrator.predict(parametric_calibrated)

    def get_timing(self):
        return self.timing



from calibration import HistogramMarginalCalibrator
from smartcal.calibration_algorithms.histogram.histogram_scaling import HistogramScalingCalibrator

from enum import Enum


class HistogramTypesEnum(Enum):
    HISTOGRAM = HistogramScalingCalibrator
    HISTOGRAMMARGINAL = HistogramMarginalCalibrator

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)



__all__ = ["HistogramCalibrator", "PlattCalibrator", "AdaptiveTemperatureScalingCalibrator", "TemperatureScalingCalibrator", 
        "IsotonicCalibrator", "MetaCalibrator", "BetaCalibrator", "DirichletCalibrator", "MatrixScalingCalibrator", "VectorScalingCalibrator", 
        "EmpiricalBinningCalibrator", "MixAndMatchCalibrator", "ProbabilityTreeCalibrator"]

def _lazy_imports():
    global HistogramCalibrator, PlattCalibrator, AdaptiveTemperatureScalingCalibrator, TemperatureScalingCalibrator, \
        IsotonicCalibrator, MetaCalibrator, BetaCalibrator, DirichletCalibrator, MatrixScalingCalibrator, VectorScalingCalibrator, \
        EmpiricalBinningCalibrator, MixAndMatchCalibrator, ProbabilityTreeCalibrator
    
    from smartcal.calibration_algorithms.histogram.histogram import HistogramCalibrator
    from smartcal.calibration_algorithms.platt.platt import PlattCalibrator
    from smartcal.calibration_algorithms.adaptive_temperature_scaling import AdaptiveTemperatureScalingCalibrator
    from smartcal.calibration_algorithms.temperature_scaling import TemperatureScalingCalibrator
    from smartcal.calibration_algorithms.isotonic import IsotonicCalibrator
    from smartcal.calibration_algorithms.meta import MetaCalibrator
    from smartcal.calibration_algorithms.beta import BetaCalibrator
    from smartcal.calibration_algorithms.dirichlet import DirichletCalibrator
    from smartcal.calibration_algorithms.matrix_scaling import MatrixScalingCalibrator
    from smartcal.calibration_algorithms.vector_scaling import VectorScalingCalibrator
    from smartcal.calibration_algorithms.empirical_binning import EmpiricalBinningCalibrator
    from smartcal.calibration_algorithms.mix_and_match import MixAndMatchCalibrator
    from smartcal.calibration_algorithms.probability_tree import ProbabilityTreeCalibrator

_lazy_imports()



from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score, log_loss
import numpy as np


def compute_classification_metrics(y_true, y_pred, y_prob=None, zero_division=0):
    """
    Compute log loss, accuracy, macro, micro, and weighted precision, recall, and F1-score.
    
    Args:
        y_true (list or array): True labels
        y_pred (list or array): Predicted labels
        y_prob (list or array, optional): Predicted probabilities for log loss calculation (required for log loss)
    
    Returns:
        dict: Dictionary containing the computed metrics
    """
    metrics = {}
    
    if y_prob is not None:
        # Ensure y_prob is a NumPy array
        y_prob = np.array(y_prob)

        # Extract unique classes from y_true
        unique_classes_true = np.unique(y_true)

        # Ensure y_prob has a valid shape
        if y_prob.ndim == 1:
            raise ValueError("y_prob should have shape (n_samples, n_classes), but has shape:", y_prob.shape)

        # Extract class indices from predictions
        unique_classes_pred = np.arange(y_prob.shape[1])

        # Get all unique classes
        all_classes = np.union1d(unique_classes_true, unique_classes_pred)

        # Compute log_loss with explicitly defined labels
        metrics["loss"] = log_loss(y_true, y_prob, labels=all_classes)
    
    metrics.update({
        'accuracy': accuracy_score(y_true, y_pred),
        'recall_macro': recall_score(y_true, y_pred, average='macro', zero_division=zero_division),
        'recall_micro': recall_score(y_true, y_pred, average='micro', zero_division=zero_division),
        'recall_weighted': recall_score(y_true, y_pred, average='weighted', zero_division=zero_division),
        'precision_macro': precision_score(y_true, y_pred, average='macro', zero_division=zero_division),
        'precision_micro': precision_score(y_true, y_pred, average='micro', zero_division=zero_division),
        'precision_weighted': precision_score(y_true, y_pred, average='weighted', zero_division=zero_division),
        'f1_macro': f1_score(y_true, y_pred, average='macro', zero_division=zero_division),
        'f1_micro': f1_score(y_true, y_pred, average='micro', zero_division=zero_division),
        'f1_weighted': f1_score(y_true, y_pred, average='weighted', zero_division=zero_division)
    })

    return metrics



from enum import Enum

from smartcal.calibration_algorithms.histogram.histogram import HistogramCalibrator
from smartcal.calibration_algorithms.platt.platt import PlattCalibrator
from smartcal.calibration_algorithms.temperature_scaling import TemperatureScalingCalibrator
from smartcal.calibration_algorithms.adaptive_temperature_scaling import AdaptiveTemperatureScalingCalibrator
from smartcal.calibration_algorithms.isotonic import IsotonicCalibrator
from smartcal.calibration_algorithms.meta import MetaCalibrator
from smartcal.calibration_algorithms.beta import BetaCalibrator
from smartcal.calibration_algorithms.dirichlet import DirichletCalibrator
from smartcal.calibration_algorithms.matrix_scaling import MatrixScalingCalibrator
from smartcal.calibration_algorithms.vector_scaling import VectorScalingCalibrator
from smartcal.calibration_algorithms.mix_and_match import MixAndMatchCalibrator
from smartcal.calibration_algorithms.empirical_binning import EmpiricalBinningCalibrator


class CalibrationAlgorithmTypesEnum(Enum):
    EMPIRICALBINNING = EmpiricalBinningCalibrator  
    ISOTONIC = IsotonicCalibrator 
    BETA = BetaCalibrator  
    TEMPERATURESCALING = TemperatureScalingCalibrator
    DIRICHLET = DirichletCalibrator
    META = MetaCalibrator
    MATRIXSCALING = MatrixScalingCalibrator
    VECTORSCALING = VectorScalingCalibrator
    PLATT = PlattCalibrator
    HISTOGRM = HistogramCalibrator
    MIXANDMATCH = MixAndMatchCalibrator
    AdaptiveTemperatureScaling = AdaptiveTemperatureScalingCalibrator

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)
    
    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"



import joblib
import pandas as pd
import numpy as np
import importlib.resources as pkg_resources
from pathlib import Path
import logging

from smartcal.meta_model.meta_model_base import BaseMetaModel
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager

try:
    import smartcal.config.resources.models as model_pkg
except ImportError:
    model_pkg = None


config_manager = ConfigurationManager()

class MetaModel(BaseMetaModel):
    """
    MetaModel loads and uses a metric-specific model and label encoder to predict the best calibration method(s)
    for a given set of input features. It supports dynamic loading of models from a package resource directory.
    """
    def __init__(self, metric: str = config_manager.metric, top_n: int = config_manager.k_recommendations):
        """
        Initialize the MetaModel with a specific metric and number of top recommendations.

        :param metric: The calibration metric to use (e.g., 'ECE', 'MCE', etc.)
        :param top_n: Number of top models to return (default from config).
        """
        super().__init__(metric=metric, top_n=top_n)
        self._model_package = model_pkg
        self.model_path = self._get_metric_model_path()
        self.label_encoder_path = self._get_metric_label_encoder_path()
        
        # Load components with validation
        self.model = self._load_component(self.model_path)
        self.label_encoder = self._load_component(self.label_encoder_path)
        
        # Validate that critical components are loaded
        self._validate_components()

    def _validate_components(self):
        """
        Validate that required components (model and label_encoder) are properly loaded.
        
        :raises RuntimeError: If critical components failed to load.
        """
        if self.model is None:
            error_msg = (
                f"Failed to load meta-model for metric '{self.metric}'. "
                f"Expected model file at: {self.model_path}. "
                f"This could be due to:\n"
                f"1. Missing model files in the package\n"
                f"2. Incorrect package installation\n"
                f"3. Missing __init__.py in models directory\n"
                f"4. Incompatible joblib version"
            )
            logging.error(error_msg)
            raise RuntimeError(error_msg)
            
        if self.label_encoder is None:
            error_msg = (
                f"Failed to load label encoder for metric '{self.metric}'. "
                f"Expected encoder file at: {self.label_encoder_path}"
            )
            logging.error(error_msg)
            raise RuntimeError(error_msg)

    def _get_model_pkg_base_path(self):
        """
        Get the base filesystem path for the models package.
        :return: Path to the models package directory.
        :raises ImportError: If the models package is not available.
        """
        if self._model_package is None:
            raise ImportError("model_pkg is not available. The models package could not be imported.")
        return model_pkg.__path__[0]

    def _get_metric_model_path(self) -> str:
        """
        Construct the path to the metric-specific model file.
        :return: Path to the model file for the selected metric.
        """
        try:
            base_path = self._get_model_pkg_base_path()
            return f"{base_path}/{self.metric}/AdaBoost.joblib"
        except Exception as e:
            logging.error(f"Failed to construct model path for metric '{self.metric}': {e}")
            return None

    def _get_metric_label_encoder_path(self) -> str:
        """
        Construct the path to the metric-specific label encoder file.
        :return: Path to the label encoder file for the selected metric.
        """
        try:
            base_path = self._get_model_pkg_base_path()
            return f"{base_path}/{self.metric}/label_encoder.joblib"
        except Exception as e:
            logging.error(f"Failed to construct label encoder path for metric '{self.metric}': {e}")
            return None

    def _load_component_fallback(self, metric: str, filename: str):
        """
        Fallback method to load components using direct file system paths.
        This is used when importlib.resources approach fails.
        
        :param metric: The metric name (e.g., 'ECE', 'MCE')
        :param filename: The filename to load (e.g., 'AdaBoost.joblib', 'label_encoder.joblib')
        :return: Loaded component or None if loading fails
        """
        try:
            # Get the directory where this module is located
            current_dir = Path(__file__).parent.parent
            models_dir = current_dir / "config" / "resources" / "models" / metric
            file_path = models_dir / filename
            
            if file_path.exists():
                component = joblib.load(str(file_path))
                logging.info(f"Successfully loaded component using fallback: {file_path}")
                return component
            else:
                logging.error(f"File does not exist at fallback path: {file_path}")
                return None
                
        except Exception as e:
            logging.error(f"Fallback loading failed for {metric}/{filename}: {str(e)}")
            return None

    def _load_component(self, path):
        """
        Load a serialized component (model or encoder) from the given path using joblib.
        :param path: Path to the .joblib file.
        :return: Loaded Python object or None if not found or error.
        """
        if path is None:
            logging.warning("Component path is None, cannot load component")
            return None
            
        # First try importlib.resources approach
        if self._model_package is not None:
            try:
                filename = Path(path).name
                subdir = Path(path).parent.name
                resource = pkg_resources.files(self._model_package).joinpath(subdir).joinpath(filename)
                
                if resource.exists():
                    with pkg_resources.as_file(resource) as f:
                        component = joblib.load(f)
                        logging.info(f"Successfully loaded component: {subdir}/{filename}")
                        return component
                else:
                    logging.warning(f"Resource does not exist: {subdir}/{filename}")
                    
            except Exception as e:
                logging.warning(f"importlib.resources approach failed: {str(e)}")
        
        # Fallback to direct file system loading
        logging.info("Attempting fallback loading method...")
        filename = Path(path).name
        metric = Path(path).parent.name
        return self._load_component_fallback(metric, filename)

    def predict_best_model(self, input_features: dict) -> list:
        """
        Predict the best calibration models for the given input features.
        :param input_features: Dictionary of input features for the meta model.
        :return: List of (class_name, normalized_probability) tuples for the top recommendations.
        """
        # Validate that model is available before prediction
        if self.model is None:
            raise RuntimeError(
                f"Meta-model for metric '{self.metric}' is not available. "
                "Please check that the package was installed correctly and all model files are present."
            )
            
        try:
            X_input = pd.DataFrame([input_features])
            y_proba = self.model.predict_proba(X_input)[0]
            
            if self.label_encoder is not None:
                class_names = self.label_encoder.classes_
            elif hasattr(self.model, 'classes_'):
                class_names = self.model.classes_
            else:
                class_names = np.arange(len(y_proba))
                
            return self._select_and_normalize(y_proba, class_names)
            
        except Exception as e:
            logging.error(f"Error during model prediction: {str(e)}")
            raise RuntimeError(f"Failed to predict best model: {str(e)}")


import numpy as np
from scipy.optimize import minimize

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.utils.timer import time_operation


class PlattScalingCalibrator(CalibratorBase):
    def __init__(self, seed=CalibratorBase.DEFAULT_SEED):
        super().__init__(seed=seed)
        self.coefficients = {}
        self.n_classes_ = None
        self.timing = {} 
        self.metadata.update({
            "calibration_type": "platt_scaling",
        })
  
    def _fit_one_class(self, logits: np.ndarray, binary_labels: np.ndarray):
        """Fit Platt scaling for a single class on log-odds."""

        def loss(params, reg_strength=1.0):
            w, b = params
            p = 1.0 / (1.0 + np.exp(-(w * logits + b)))
            p = np.clip(p, 1e-15, 1 - 1e-15)
            # Negative log-likelihood
            nll = -np.sum(binary_labels * np.log(p) + (1 - binary_labels) * np.log(1 - p))
            # L2 penalty on w (not usually on b)
            penalty = 0.5 * reg_strength * (w ** 2)
            return nll + penalty

        initial_params = [1.0, 0.0]
        result = minimize(loss, initial_params, method='L-BFGS-B')
        if not result.success:
            self.logger.warning(f"Optimization failed: {result.message}")
            return initial_params  # fallback
        return result.x

    @time_operation
    def fit(self, predictions: np.ndarray, ground_truth: np.ndarray):
        """
        Fit Platt scaling parameters. We treat `predictions` as probabilities, so we
        convert them to log-odds. For binary classification, we only calibrate the
        "positive" column. For multi-class, we do one-vs-rest.
        """
        predictions, ground_truth = super().validate_inputs(predictions, ground_truth)

        self.n_classes_ = predictions.shape[1]
        self.coefficients = {}

        # Log dataset info
        self.metadata["dataset_info"].update({
            "n_samples": len(predictions),
            "n_classes": self.n_classes_
        })
        self.logger.info(f"PlattScaling.fit started with {len(predictions)} samples, {self.n_classes_} classes")

        # === Binary special case: calibrate only the positive column. ===
        if self.n_classes_ == 2:
            # The "positive" column is index 1
            p = np.clip(predictions[:, 1], 1e-15, 1 - 1e-15)
            logits = np.log(p / (1 - p))

            # Binary labels for the "positive" class
            binary_labels = (ground_truth == 1).astype(int)

            self.coefficients[1] = self._fit_one_class(logits, binary_labels)

        else:
            # === Multi-class: one-vs-rest approach ===
            for class_idx in range(self.n_classes_):
                binary_labels = (ground_truth == class_idx).astype(int)

                p = np.clip(predictions[:, class_idx], 1e-15, 1 - 1e-15)
                logits = np.log(p / (1.0 - p))

                self.coefficients[class_idx] = self._fit_one_class(logits, binary_labels)

        self.fitted = True

        # Store coefficients
        self.metadata["params"]["coefficients"] = {str(k): list(v) for k, v in self.coefficients.items()}
        self.logger.info(f"PlattScaling fitted coefficients: {self.metadata['params']['coefficients']}")

    @time_operation
    def predict(self, test_data: np.ndarray) -> np.ndarray:
        """
        Apply Platt scaling to test data, assumed to be probabilities. For binary,
        we only calibrate the positive column. For multi-class, we calibrate
        each column (one-vs-rest) and then re-normalize.
        """
        if not self.fitted:
            raise RuntimeError("Calibrator not fitted. Call `fit` first.")

        n_samples, n_cols = test_data.shape
        if n_cols != self.n_classes_:
            raise ValueError("Number of columns in `test_data` differs from training phase.")

        # === Binary special case ===
        if self.n_classes_ == 2:
            calibrated_probs = np.zeros((n_samples, 2))
            w, b = self.coefficients[1]

            # Convert the positive column to log-odds
            p = np.clip(test_data[:, 1], 1e-15, 1 - 1e-15)
            logits = np.log(p / (1 - p))

            # Sigmoid(w * logit + b)
            pos_calibrated = 1.0 / (1.0 + np.exp(-(w * logits + b)))
            neg_calibrated = 1.0 - pos_calibrated

            calibrated_probs[:, 0] = neg_calibrated
            calibrated_probs[:, 1] = pos_calibrated
            return calibrated_probs

        else:
            # === Multi-class scenario: one-vs-rest + re-normalize ===
            calibrated_probs = np.zeros((n_samples, self.n_classes_))
            for class_idx in range(self.n_classes_):
                w, b = self.coefficients[class_idx]
                p = np.clip(test_data[:, class_idx], 1e-15, 1 - 1e-15)
                logits = np.log(p / (1.0 - p))
                calibrated_probs[:, class_idx] = 1.0 / (1.0 + np.exp(-(w * logits + b)))

            # Re-normalize row-wise
            row_sums = calibrated_probs.sum(axis=1, keepdims=True)
            calibrated_probs /= np.clip(row_sums, 1e-15, None)
            return calibrated_probs
               
    def get_timing(self):
        """Return the timing information for fit and predict operations."""
        return self.timing

    def get_timing(self):
        return self.timing


import os

from smartcal.config.configuration_manager.configuration_manager_base import ConfigurationManagerBase
from smartcal.config.configuration_manager.configuration_manager_factory import ConfigManagerFactory
from smartcal.config.enums.config_reader_enum import ConfigReaderEnum


class ConfigurationManager(ConfigurationManagerBase):
    config_reader = ConfigReaderEnum.get_value(os.getenv("CONFIG_MANAGER", "env"))
    config_manager = ConfigManagerFactory(config_reader=config_reader)
    config = config_manager.get_config()

    @property
    def n_bins(self):
        return int(self.config('N_BIN', default=10))
    
    @property
    def n_bins_calibraion_curve(self):
        return int(self.config('N_BIN_CAL_CURVE', default=5))

    @property
    def device(self):
        return str(self.config('DEVICE'))
    
    @property
    def patience(self):
        return int(self.config('PATIENCE', default=5))
    
    @property
    def min_lr_img(self):
        return float(self.config('MIN_LR_IMG'))
    
    @property
    def max_lr_img(self):
        return float(self.config('MAX_LR_IMG', default=5))
    
    @property
    def min_delta_early_stopper_img(self):
        return float(self.config('MIN_DELTA_EARLY_STOPPER_IMG', default=5))
    
    @property
    def num_itr_early_stopper_img(self):
        return int(self.config('NUM_ITR_EARLY_STOPPER_IMG', default=5))
    
    @property    
    def language_transformer_lr_min(self):
        return float(self.config('LANGUAGE_TRANSFORMER_LR_MIN'))
    
    @property    
    def language_transformer_lr_max(self):
        return float(self.config('LANGUAGE_TRANSFORMER_LR_MAX'))
    
    @property    
    def language_fasttext_lr_min(self):
        return float(self.config('LANGUAGE_FASTTEXT_LR_MIN'))
    
    @property    
    def language_fasttext_lr_max(self):
        return float(self.config('LANGUAGE_FASTTEXT_LR_MAX'))
    
    @property    
    def language_lr_finder_epochs(self):
        return int(self.config('LANGUAGE_LR_FINDER_EPOCHS'))
    
    @property    
    def language_train_epochs(self):
        return int(self.config('LANGUAGE_TRAIN_EPOCHS'))
    
    @property    
    def language_patience(self):
        return int(self.config('LANGUAGE_PATIENCE'))
    
    @property    
    def language_min_delta(self):
        return float(self.config('LANGUAGE_MIN_DELTA'))
    
    @property    
    def language_batch_size(self):
        return int(self.config('LANGUAGE_BATCH_SIZE'))
    
    @property    
    def language_monitor_metric(self):
        return str(self.config('LANGUAGE_MONITOR_METRIC'))
    
    @property
    def n_bin_cal(self):
        return int(self.config('N_BIN_CAL', default=10))
    
    @property
    def beta_model(self):
        return self.config('BETA_MODEL', 'abm')
    
    @property
    def random_seed(self):
        return int(self.config('RANDOM_SEED', default=42))

    @property
    def n_bin_metrics(self):
        return int(self.config('N_BIN_METRICS'))

    @property
    def conf_threshold(self):
        return float(self.config('CONF_THRESHOLD'))
    
    @property
    def conf_thresholds_list(self):
        thresholds = self.config('CONF_THRESHOLDS_LIST')
        threshold_strings = thresholds.strip('[]').split(',')
        return [float(threshold.strip()) for threshold in threshold_strings]

    @property
    def split_ratios(self):
        split_ratios_str = self.config('SPLIT_RATIOS', default='(0.7,0.15,0.15)')
        return tuple(map(float, split_ratios_str.strip('()').split(',')))

    @property
    def split_ratios_images(self):
        split_ratios_images_str = self.config('SPLIT_RATIOS_IMAGES', default='(0.8,0.2)')
        return tuple(map(float, split_ratios_images_str.strip('()').split(',')))

    @property
    def batch_size(self):
        return int(self.config('BATCH_SIZE', default=32))

    @property
    def img_size(self):
        return int(self.config('IMG_SIZE', default=224))

    @property
    def tokenizer_max_length(self):
        return int(self.config('TOKENIZER_MAX_LENGTH', default=512))

    @property
    def bert_tokenizer(self):
        return str(self.config('BERT_TOKENIZER', default='bert-base-uncased'))

    @property
    def config_img(self):
        return str(self.config('CONFIG_IMG'))

    @property
    def config_tabular(self):
        return str(self.config('CONFIG_TABULAR'))

    @property
    def config_language(self):
        return str(self.config('CONFIG_LANGUAGE'))
    
    @property
    def lr_tempscaling(self):
        return float(self.config('LR_TEMPSCALING'))

    @property
    def max_iter_tempscaling(self):
        return int(self.config('MAX_ITER_TEMPSCALING'))
    @property
    def lr_vectorscaling(self):
        return float(self.config('LR_VECTORSCALING'))

    @property
    def max_iter_vectorscaling(self):
        return int(self.config('MAX_ITER_VECTORSCALING'))

    @property
    def lr_matrixscaling(self):
        return float(self.config('LR_MATRIXSCALING'))

    @property
    def max_iter_matrixscaling(self):
        return int(self.config('MAX_ITER_MATRIXSCALING'))
 
    @property
    def initial_T(self):
        return float(self.config('INITIAL_T'))  
    
    @property
    def n_bin_meta_features(self):
        return int(self.config('N_BIN_META_FEATURES')) 

    @property
    def meta_data_file(self):
        return str(self.config('META_DATA_FILE'))  

    @property    
    def lr_dirichlet(self):
        return float(self.config('LR_DIRICHLET'))

    @property    
    def max_iter_dirichlet(self):
        return int(self.config('MAX_ITER_DIRICHLET'))    
    
    @property
    def epochs(self):
        return int(self.config('EPOCHS', default=5))   
    
    @property
    def learning_rate(self):
        return float(self.config('LEARNING_RATE', default=0.001))

    @property    
    def probability_tree_max_depth(self):
        return int(self.config('PROBABILITYTREE_MAX_DEPTH', default=5))

    @property    
    def probability_tree_min_samples_leaf(self):
        return int(self.config('PROBABILITYTREE_MIN_SAMPLES_LEAF', default=15))

    @property    
    def meta_alpha(self):
        return float(self.config('META_ALPHA', default=0.1))

    @property    
    def meta_acc(self):
        return float(self.config('META_ACC', default=0.85))
    
    @property    
    def meta_default_constraint(self):
        return str(self.config('META_DEFAULT_CONSTRAINT')) # can be ALPHA or ACC
    
    @property    
    def platt_default_type(self):
        return str(self.config('PLATT_DEFAULT_TYPE')) # can be PLATT, PLATTBINNER or PLATTBINNERMARGINAL
    
    @property    
    def histogram_default_type(self):
        return str(self.config('HISTOGRAM_DEFAULT_TYPE')) # can be HISTOGRAM or HISTOGRAMMARGINAL

    @property
    def n_confidence_bins(self):
        return int(self.config('N_CONFIDENCE_BINS', default=10))

    @property
    def n_entropy_bins(self):
        return int(self.config('N_ENTROPY_BINS', default=10))
    
    # FOR HYPERPARAMETER TUNING
    @property    
    def beta_params_tune(self):
        param_types = self.config('BETA_PARAMS_TUNE')
        param_types = param_types.strip("'[]'")
        return [param.strip().strip("'") for param in param_types.split(',')]

    @property    
    def cal_n_bins_tune(self):
        n_bins = self.config('CAL_NUM_BINS_TUNE')
        n_bins = n_bins.strip("'[]'")
        return [int(bin_value.strip()) for bin_value in n_bins.split(',')]
    
    @property    
    def cal_lr_tune(self):
        lr_values = self.config('CAL_LR_TUNE')
        lr_values = lr_values.strip("'[]'")
        return [float(lr.strip()) for lr in lr_values.split(',')]
    
    
    @property    
    def cal_max_iter_tune(self):
        max_iter = self.config('CAL_MAX_ITER_TUNE')
        max_iter = max_iter.strip("'[]'")
        return [int(iter_value.strip()) for iter_value in max_iter.split(',')]
    
    @property    
    def meta_calibrator_calibrator_type_tune(self):
        calibrator_types = self.config('META_CALIBRATOR_CALIBRATOR_TYPE_TUNE')
        calibrator_types = calibrator_types.strip("'[]'")
        return [type.strip().strip("'") for type in calibrator_types.split(',')]
    
    @property    
    def meta_alpha_tune(self):
        alpha_values = self.config('META_ALPHA_TUNE')
        alpha_values = alpha_values.strip("'[]'")
        return [float(alpha.strip()) for alpha in alpha_values.split(',')]
    
    @property    
    def meta_acc_tune(self):
        acc_values = self.config('META_ACC_TUNE')
        acc_values = acc_values.strip("'[]'")
        return [float(acc.strip()) for acc in acc_values.split(',')]
    
    @property    
    def probability_tree_max_depth_tune(self):
        max_depth_values = self.config('PROBABILITY_TREE_MAX_DEPTH_TUNE')
        max_depth_values = max_depth_values.strip("'[]'")
        return [None if value.strip() == 'None' else int(value.strip()) for value in max_depth_values.split(',')]
    
    @property    
    def probability_tree_min_samples_leaf_tune(self):
        min_samples_leaf = self.config('PROBABILITY_TREE_MIN_SAMPLES_LEAF_TUNE')
        min_samples_leaf = min_samples_leaf.strip("'[]'")
        return [int(value.strip()) for value in min_samples_leaf.split(',')]
    
    @property    
    def temperature_scaling_initial_t_tune(self):
        initial_t_values = self.config('TEMPERATURE_SCALING_INITIAL_T_TUNE')
        initial_t_values = initial_t_values.strip("'[]'")
        return [float(value.strip()) for value in initial_t_values.split(',')]
    
    @property    
    def histogram_calibrator_calibrator_type_tune(self):
        calibrator_types = self.config('HISTOGRAM_CALIBRATOR_CALIBRATOR_TYPE_TUNE')
        calibrator_types = calibrator_types.strip("'[]'")
        return [type.strip().strip("'") for type in calibrator_types.split(',')]
    
    @property    
    def platt_calibrator_calibrator_type_tune(self):
        calibrator_types = self.config('PLATT_CALIBRATOR_CALIBRATOR_TYPE_TUNE')
        calibrator_types = calibrator_types.strip("'[]'")
        return [type.strip().strip("'") for type in calibrator_types.split(',')]
       
    @property    
    def adaptive_temp_mode_tune(self):
        adapt_temp_modes = self.config('ADAPT_TEMP_SCALING_MODES_TUNE')
        adapt_temp_modes = adapt_temp_modes.strip("'[]'")
        return [type.strip().strip("'") for type in adapt_temp_modes.split(',')]
    
    @property    
    def adaptive_temp_bins_tune(self):
        temp_bins = self.config('ADAPT_TEMP_SCALING_BINS_TUNE')
        temp_bins = temp_bins.strip("'[]'")
        return [int(value.strip()) for value in temp_bins.split(',')]
     
    @property    
    def adaptive_temp_lr_tune(self):
        return float(self.config('ADAPT_TEMP_SCALING_lR_TUNE', default=0.1))

    @property    
    def adaptive_temp_max_itr_tune(self):
        return int(self.config('ADAPT_TEMP_SCALING_MAX_ITR_TUNE', default=100))
    
    @property    
    def mix_match_parametric_tune(self):
        parametric_cals = self.config('MIX_MATCH_PARAMETRIC')
        parametric_cals_values = parametric_cals.strip("'[]'")
        return [type.strip().strip("'") for type in parametric_cals_values.split(',')]
     
    @property    
    def mix_match_nonparametric_tune(self):
        nonparametric_cals = self.config('MIX_MATCH_NONPARAMETRIC')
        nonparametric_cals_values = nonparametric_cals.strip("'[]'")
        return [type.strip().strip("'") for type in nonparametric_cals_values.split(',')]
    
    ## For DB
    @property
    def db_host(self):
        return str(self.config('DB_HOST'))

    @property
    def db_port(self):
        return int(self.config('DB_PORT'))

    @property
    def db_name(self):
        return str(self.config('DB_NAME'))

    @property
    def db_user(self):
        return str(self.config('DB_USER'))

    @property
    def db_password(self):
        return str(self.config('DB_PASSWORD'))
    
    @property    
    def cal_tune_kfolds(self):
        return int(self.config('CAL_KFOLDS'))
    
    # Remote machine access
    @property
    def ssh_enabled(self):
        return str(self.config('SSH_ENABLED')).strip().lower() == "true"

    @property
    def ssh_host(self):
        return str(self.config('SSH_HOST'))

    @property
    def ssh_port(self):
        return int(self.config('SSH_PORT'))

    @property
    def ssh_user(self):
        return str(self.config('SSH_USER'))

    @property
    def ssh_key_path(self):
        return str(self.config('SSH_KEY_PATH'))

    @property
    def remote_bind_port(self):
        return int(self.config('REMOTE_BIND_PORT'))

    @property
    def local_bind_port(self):
        return int(self.config('LOCAL_BIND_PORT'))

    @property
    def baseline_results(self):
        return str(self.config('BASELINE_RESULTS'))

    @property
    def k_recommendations(self):
        return int(self.config('K_RECOMMENDATIONS', default=5))

    @property    
    def bayes_lr_tune(self):
        lr_values = self.config('BAYES_LR')
        lr_values = lr_values.strip("'()")  # Remove parentheses
        return tuple(float(value.strip()) for value in lr_values.split(','))  # Convert to float and return as tuple

    @property    
    def bayes_max_iter_tune(self):
        max_itr_values = self.config('BAYES_MAX_ITR')
        max_itr_values = max_itr_values.strip("'()")  # Remove parentheses
        return tuple(int(value.strip()) for value in max_itr_values.split(','))  # Convert to integer and return as tuple

    @property    
    def bayes_temp_tune(self):
        temp_values = self.config('BAYES_TEMP')
        temp_values = temp_values.strip("'()")  # Remove parentheses
        return tuple(float(value.strip()) for value in temp_values.split(','))  # Convert to float and return as tuple

    @property    
    def bayes_alpha_tune(self):
        alpha_values = self.config('BAYES_ALPHA')
        alpha_values = alpha_values.strip("'()")  # Remove parentheses
        return tuple(float(value.strip()) for value in alpha_values.split(','))  # Convert to float and return as tuple

    @property    
    def bayes_acc_tune(self):
        acc_values = self.config('BAYES_ACC')
        acc_values = acc_values.strip("'()")  # Remove parentheses
        return tuple(float(value.strip()) for value in acc_values.split(','))  # Convert to float and return as tuple
    
    @property
    def bayes_num_bins_tune(self):
        bins_values = self.config('BAYES_NUM_BINS')
        bins_values = bins_values.strip("'()")  # Remove parentheses
        return tuple(float(value.strip()) for value in bins_values.split(','))  # Convert to float and return as tuple
    
    @property
    def meta_models_path(self):
        return str(self.config('META_MODELS_PATH'))

    @property
    def meta_model_type(self):
        """Get the currently selected meta model type"""
        return str(self.config('META_MODEL_TYPE', default='xgb')).lower()

    @property
    def meta_model_version(self):
        return str(self.config('META_MODEL_VERSION'))

    @property
    def n_splits(self):
        return int(self.config('N_SPLITS', default=5))

    @property
    def results_dir(self):
        return str(self.config('RESULTS_DIR'))
    
    @property
    def meta_model_k(self):
        return int(self.config('META_MODEL_K', default=5))
    
    @property
    def num_itr_eval(self):
        num_itr_values = self.config('NUM_ITR_EVAL')
        num_itr_values = num_itr_values.strip("'[]'")
        return [int(value.strip()) for value in num_itr_values.split(',')] 
    
    @property
    def metric(self):
        return str(self.config('METRIC'))

    @property
    def supported_metrics(self):
        supported_metrics = self.config('SUPPORTED_METRICS')
        supported_metrics_values = supported_metrics.strip("'[]'")
        return [type.strip().strip("'") for type in supported_metrics_values.split(',')]
    
    @property
    def logging(self):
        return str(self.config('LOGGING')).strip().lower() == "true"

    @property
    def meta_label_encoder_path(self):
        return str(self.config('META_LABEL_ENCODER_PATH'))

    @property
    def insights_folder_path(self):
        return str(self.config('INSIGHTS_FOLDER_PATH'))

    @property
    def all_class_distributions_file(self):
        return str(self.config('ALL_CLASS_DISTRIBUTIONS_FILE'))



import logging

from classifiers.tabular_classifier import TabularClassifier
from classifiers.image_classifier import ImageClassifier
from classifiers.language_classifier import LanguageClassifier
from smartcal.config.enums.language_models_enum import LanguageModelsEnum
from smartcal.config.enums.dataset_types_enum import DatasetTypesEnum


class ModelCache:
    """
    Singleton class that implements a caching mechanism for machine learning models.
    Ensures that models are only initialized once and reused when needed.
    """
    
    # Class-level variables for singleton pattern and model storage
    _instance = None  # Singleton instance
    _models = {}     # Dictionary to store initialized models using tuple keys
    logger = logging.getLogger(__name__)
    
    @classmethod
    def get_model(cls, model_enum, task_type, **kwargs):
        """
        Get or create a model instance based on the specified parameters.
        
        Args:
            model_enum: Enum representing the specific model type
            task_type: Type of task (TABULAR, IMAGE, or LANGUAGE)
            **kwargs: Additional arguments including:
                - seed: Random seed for reproducibility
                - num_classes: Number of classes for classification
                - device: Device to run the model on (CPU/GPU)
                
        Returns:
            Initialized model instance from cache or newly created
            
        Raises:
            ValueError: If task type is unsupported or model enum is invalid
        """
        try:
            # Create a unique key for the model using a tuple of its identifying characteristics
            model_key = (model_enum, task_type, kwargs.get('seed', None), 
                        kwargs.get('num_classes', None), kwargs.get('device', None))
            
            # Check if model exists in cache
            if model_key not in cls._models:
                cls.logger.info(f"Initializing new {task_type} model: {model_enum}")
                
                # Initialize appropriate classifier based on task type
                if task_type == DatasetTypesEnum.TABULAR:
                    # Initialize tabular classifier (e.g., Random Forest, XGBoost)
                    cls._models[model_key] = TabularClassifier(
                        model_enum=model_enum,
                        seed=kwargs.get('seed')
                    )
                elif task_type == DatasetTypesEnum.IMAGE:
                    # Initialize image classifier (e.g., MOBILENET)
                    cls._models[model_key] = ImageClassifier(
                        model_enum=model_enum,
                        num_classes=kwargs.get('num_classes'),
                        device=kwargs.get('device'),
                        seed=kwargs.get('seed')
                    )
                elif task_type == DatasetTypesEnum.LANGUAGE:
                    # Validate that model_enum is appropriate for language tasks
                    if not isinstance(model_enum, LanguageModelsEnum):
                        raise ValueError(f"Expected LanguageModelsEnum for language task, got {type(model_enum)}")
                    
                    # Initialize language classifier (e.g., BERT, GPT)
                    cls._models[model_key] = LanguageClassifier(
                        model_enum=model_enum,
                        seed=kwargs.get('seed')
                    )
                else:
                    raise ValueError(f"Unsupported task type: {task_type}")
            else:
                # Model found in cache
                cls.logger.info(f"Retrieved cached {task_type} model: {model_enum}")
            
            # Return either the cached or newly created model
            return cls._models[model_key]
                
        except Exception as e:
            # Log any errors during model initialization
            cls.logger.error(f"Error in model initialization: {e}")
            raise

    @classmethod
    def clear_cache(cls):
        """
        Clear all cached models from memory.
        Useful for freeing up memory or resetting the cache state.
        """
        cls._models.clear()
        cls.logger.info("Model cache cleared")


from abc import ABC, abstractmethod
from decouple import Config, RepositoryEnv
import os
from pathlib import Path

from smartcal.config.enums.config_reader_enum import ConfigReaderEnum


class ConfigReader(ABC):
    @abstractmethod
    def get_config(self) -> Config:
        pass

    @abstractmethod
    def set_config(self, key: str, value: str):
        pass


class EnvConfigReader(ConfigReader):
    def __init__(self):
        __base_dir = Path(__file__).resolve().parent.parent.parent
        __file_name = f'config-dev.env'
        self.full_path = os.path.join(__base_dir, 'config', 'resources', 'conf', __file_name)

    def get_config(self) -> Config:
        return Config(repository=RepositoryEnv(self.full_path))

    def set_config(self, key: str, value: str):
        # Read all lines from the .env file
        if os.path.exists(self.full_path):
            with open(self.full_path, "r") as f:
                lines = f.readlines()
        else:
            lines = []

        # Modify the content, updating the key if it exists, or appending if not
        updated = False
        for i, line in enumerate(lines):
            if line.startswith(f"{key}"):
                lines[i] = f"{key} = '{value}'\n"
                updated = True
                break

        if not updated:
            raise KeyError(f"The configuration key '{key}' was not found in the .env file.")

        # Write the entire updated content back to the .env file, replacing existing content
        with open(self.full_path, "w") as f:
            f.writelines(lines)


class ConfigManagerFactory:
    def __new__(cls, config_reader: ConfigReader.__name__ = ConfigReaderEnum.env.value):
        _factory_localizer = {subclass.__name__: subclass for subclass in ConfigReader.__subclasses__()}
        subclass = _factory_localizer.get(config_reader)
        return subclass()



from typing import Dict, Any

from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager


config_manager = ConfigurationManager()

lr = config_manager.bayes_lr_tune
max_itr = config_manager.bayes_max_iter_tune
temp =  config_manager.bayes_temp_tune
alpha = config_manager.bayes_alpha_tune
acc = config_manager.bayes_acc_tune
num_bins = config_manager.bayes_num_bins_tune

# Parameter mappings for categorical variables
param_mappings: Dict[str, Dict[str, Any]] = {
    'BETA': {
        'model_type': {'abm': 0, 'am': 1, 'ab': 2}
    },
    'META': {
        'calibrator_type': {'ALPHA': 0, 'ACC': 1}
    },
    'PLATT': {
        'calibrator_type': {'PLATT': 0, 'PLATTBINNER': 1, 'PLATTBINNERMARGINAL': 2}
    },
    'HISTOGRM': {
        'calibrator_type': {'HISTOGRAM': 0, 'HISTOGRAMMARGINAL': 1}
    },
    'AdaptiveTemperatureScaling': {
        'mode': {'linear': 0, 'entropy': 1, 'hybrid': 2}
    },
    'MIXANDMATCH': {
        'parametric_calibrator': {
            'TemperatureScalingCalibrator': 0,
            'PlattCalibrator': 1,
            'VectorScalingCalibrator': 2,
            'MatrixScalingCalibrator': 3,
            'BetaCalibrator': 4,
            'MetaCalibrator': 5,
            'DirichletCalibrator': 6,
            'AdaptiveTemperatureScalingCalibrator': 7
        },
        'nonparametric_calibrator': {
            'IsotonicCalibrator': 0,
            'EmpiricalBinningCalibrator': 1,
            'HistogramCalibrator': 2
        }
    }
}

# Parameter spaces for optimization (using continuous ranges)
param_spaces: Dict[str, Dict[str, Any]] = {
    'EMPIRICALBINNING': {
        'n_bins': num_bins # Will be rounded to nearest integer
    },
    'TEMPERATURESCALING': {
        'initial_T': temp,
        'lr_tempscaling': lr,
        'max_iter_tempscaling': max_itr  # Will be rounded
    },
    'ISOTONIC': {},
    'BETA': {
        'model_type': (0, 2)  # Will be mapped to categorical
    },
    'DIRICHLET': {
        'lr': lr,
        'max_iter': max_itr
    },
    'META': {
        'calibrator_type': (0, 1),
        'alpha': alpha,
        'acc': acc
    },
    'MATRIXSCALING': {
        'lr': lr,
        'max_iter': max_itr
    },
    'VECTORSCALING': {
        'lr': lr,
        'max_iter': max_itr
    },
    'PLATT': {
        'calibrator_type': (0, 2),
        'num_bins': num_bins
    },
    'HISTOGRM': {
        'calibrator_type': (0, 1),
        'num_bins': num_bins
    },
    'MIXANDMATCH': {
        'parametric_calibrator': (0, 7),
        'nonparametric_calibrator': (0, 2)
    },
    'AdaptiveTemperatureScaling': {
        'initial_T': temp,
        'lr_tempscaling': lr,
        'max_iter_tempscaling': max_itr,
        'mode': (0, 2),
        'confidence_bins': num_bins,
        'entropy_bins': num_bins
    }
}



__all__ = ["ECE", "MCE", "ConfECE", "calculate_brier_score", "calculate_calibration_curve"]

def _lazy_imports():
    global ECE, MCE, ConfECE, calculate_brier_score, calculate_calibration_curve
    
    from smartcal.metrics.ece import ECE
    from smartcal.metrics.mce import MCE
    from smartcal.metrics.conf_ece import ConfECE
    from smartcal.metrics.brier_score import calculate_brier_score
    from smartcal.metrics.calibration_curve import calculate_calibration_curve

_lazy_imports()



import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import KBinsDiscretizer

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.utils.timer import time_operation  # For measuring execution time


config_manager = ConfigurationManager()

class PlattBinnerScalingCalibrator(CalibratorBase):
    def __init__(self, num_bins=config_manager.n_bin_cal, seed=CalibratorBase.DEFAULT_SEED):
        """
        Initialize the Platt Binner Scaling calibrator.
        :param num_bins: Number of bins to use in the binning step.
        """
        super().__init__(seed)
        self.num_bins = num_bins
        self.platt_models = None
        self.binners = None
        self.bin_means = None
        self.n_classes = None
        self.metadata.update({"calibration_type": "platt_binner_scaling"})  # Store calibration type
        self.metadata["params"].update({"num_bins": self.num_bins})  
        self.timing = {}


    @time_operation
    def fit(self, logits, labels):
        """
        Fit the Platt Scaling model and apply binning for multi-class calibration.
        :param logits: (n_samples, n_classes) array of logits (pre-softmax scores).
        :param labels: (n_samples,) array of true labels.
        """
        logits, labels = super().validate_inputs(logits, labels)  # Ensure input consistency
        logits = np.array(logits)
        labels = np.array(labels)
        n_samples, self.n_classes = logits.shape

        self.platt_models = []
        self.binners = []
        self.bin_means = []

        # Log dataset info
        self.metadata["dataset_info"].update({"n_samples": n_samples, "n_classes": self.n_classes})

        # One-vs-All Platt Scaling + Binning
        for class_idx in range(self.n_classes):
            # Convert labels to binary: is this class or not?
            binary_labels = (labels == class_idx).astype(int)

            # Train logistic regression for Platt Scaling
            model = LogisticRegression(solver="lbfgs")
            model.fit(logits[:, class_idx].reshape(-1, 1), binary_labels)
            self.platt_models.append(model)

            # Compute Platt-scaled probabilities
            platt_probs = model.predict_proba(logits[:, class_idx].reshape(-1, 1))[:, 1]

            # Train binning model (Equal-Width Bins)
            try:
                binner = KBinsDiscretizer(n_bins=self.num_bins, encode="ordinal", strategy="uniform")
                bin_indices = binner.fit_transform(platt_probs.reshape(-1, 1)).astype(int).flatten()
            except ValueError as e:
                self.logger.error(f"Binning failed for class {class_idx}: {e}")
                raise RuntimeError("Binning encountered an issue. Try reducing `num_bins`.")

            # Compute actual bin means from training data
            bin_means = np.zeros(self.num_bins)
            for b in range(self.num_bins):
                in_bin = platt_probs[bin_indices == b]
                bin_means[b] = np.mean(in_bin) if len(in_bin) > 0 else 0.5  # Default to 0.5 if bin is empty

            self.binners.append(binner)
            self.bin_means.append(bin_means)

        self.fitted = True

    @time_operation
    def predict(self, logits):
        """
        Apply Platt Scaling + Binning to obtain calibrated probabilities.
        :param logits: (n_samples, n_classes) array of logits.
        :return: (n_samples, n_classes) array of calibrated probabilities.
        """
        if not self.fitted:
            raise RuntimeError("Calibrator not fitted. Call `fit` first.")

        logits = np.array(logits)

        # Ensure input shape matches the number of classes
        if logits.shape[1] != self.n_classes:
            raise ValueError(f"Mismatch in number of classes: expected {self.n_classes}, got {logits.shape[1]}.")

        n_samples = logits.shape[0]
        calibrated_probs = np.zeros((n_samples, self.n_classes))

        for class_idx in range(self.n_classes):
            # Apply Platt Scaling
            platt_probs = self.platt_models[class_idx].predict_proba(logits[:, class_idx].reshape(-1, 1))[:, 1]

            # Apply Binning Calibration
            try:
                bin_indices = self.binners[class_idx].transform(platt_probs.reshape(-1, 1)).astype(int).flatten()
            except ValueError as e:
                raise RuntimeError(f"Binning failed for class {class_idx}: {e}")

            # Ensure bin indices are within range
            bin_indices = np.clip(bin_indices, 0, self.num_bins - 1)

            # Assign bin means
            calibrated_probs[:, class_idx] = self.bin_means[class_idx][bin_indices]

        # Normalize probabilities to ensure they sum to 1
        row_sums = calibrated_probs.sum(axis=1, keepdims=True)
        calibrated_probs /= np.clip(row_sums, 1e-15, None)  # Avoid division by zero

        return calibrated_probs

    def get_timing(self):
        return self.timing


import numpy as np

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.utils.timer import time_operation
from smartcal.config.enums.platt_calibrator_types_enum import PlattTypesEnum


config_manager = ConfigurationManager()

class PlattCalibrator(CalibratorBase):
    """
    Implements Platt scaling for multi-class calibration.
    """
    def __init__(self, calibrator_type=config_manager.platt_default_type, num_bins=config_manager.n_bin_cal, seed=CalibratorBase.DEFAULT_SEED):
        super().__init__(seed)
        
        if not isinstance(calibrator_type, str):
            raise ValueError(f"Invalid calibrator type: {calibrator_type}. Must be a string.")

        try:
            self.calibrator_type = PlattTypesEnum[calibrator_type.upper()]
        except KeyError:
            raise ValueError(f"Unknown calibrator type: {calibrator_type}. Available options: {list(PlattTypesEnum.__members__.keys())}")

        self.num_bins = num_bins
        self.calibrator = None
        self.fitted = False
        self.timing = {}

    @time_operation
    def fit(self, predictions, ground_truth):
        """
        Trains the Platt scaling calibrator.

        Args:
            predictions (np.ndarray): Model logits of shape (n_samples, n_classes).
            ground_truth (np.ndarray): True labels of shape (n_samples,).
        """
        self.set_random_seed(self.seed)  # Ensure full determinism

        predictions, ground_truth = self.validate_inputs(predictions, ground_truth)

        self.metadata["dataset_info"].update({
            "n_samples": predictions.shape[0],
            "n_classes": predictions.shape[1],
        })
        self.metadata["params"].update({
            "Platt Calibration Type": self.calibrator_type.name,
            "num_bins": self.num_bins
        })

        num_samples = len(ground_truth)  # Required for instantiation

        if self.calibrator_type == PlattTypesEnum.PLATT:
            self.calibrator = self.calibrator_type.value(seed=self.seed)
            self.calibrator.fit(predictions, ground_truth)
        elif self.calibrator_type == PlattTypesEnum.PLATTBINNER:
            self.calibrator = self.calibrator_type.value(num_bins=self.num_bins, seed=self.seed)
            self.calibrator.fit(predictions, ground_truth)
        else:
            # Dynamically instantiate the correct calibrator model with both arguments
            self.calibrator = self.calibrator_type.value(num_calibration=num_samples, num_bins=self.num_bins)  
            self.calibrator.train_calibration(predictions, ground_truth)

        self.fitted = True

    @time_operation
    def predict(self, test_data):
        """
        Applies Platt scaling to logits.

        Args:
            test_data (np.ndarray): Logits of shape (n_samples, n_classes).

        Returns:
            np.ndarray: Calibrated probabilities of shape (n_samples, n_classes).
        """
        if not self.fitted:
            raise RuntimeError("Model must be fitted before prediction.")

        if not isinstance(test_data, np.ndarray):
            raise ValueError("Test data must be a NumPy array.")

        if test_data.ndim != 2:
            raise ValueError(f"Test data must be a 2D NumPy array, but got shape {test_data.shape}.")
        
        # Apply the correct calibration method based on PlattTypesEnum
        if self.calibrator_type in {PlattTypesEnum.PLATT, PlattTypesEnum.PLATTBINNER}:
            calibrated_prob = self.calibrator.predict(test_data)
        else:
            calibrated_prob = self.calibrator.calibrate(test_data)

        # Normalize probabilities to sum to 1
        row_sums = calibrated_prob.sum(axis=1, keepdims=True)
        calibrated_prob /= np.clip(row_sums, 1e-15, None)  # Avoid division by zero

        return calibrated_prob

    def get_timing(self):
        return self.timing


import numpy as np
import logging

from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager


logger = logging.getLogger(__name__)
config_manager = ConfigurationManager()

def calculate_calibration_curve(y_true, y_prob, n_bins = config_manager.n_bins_calibraion_curve):
    """
    Calculate calibration curve data for multiclass classification.
    
    Parameters:
    -----------
    y_true : array-like
        True labels (can be binary or multiclass)
    y_prob : array-like
        Predicted probabilities for all classes
        Shape should be (n_samples, n_classes)
    n_bins : int
        Number of bins for the calibration curve
        
    Returns:
    --------
    tuple or None
        - mean_predicted_probs: array of mean predicted probabilities per bin
        - true_probs: array of true probabilities per bin
        - bin_counts: array of sample counts per bin
        Returns None if inputs are invalid.
    """
    try:
        y_true = np.array(y_true).astype(int)
        y_prob = np.array(y_prob)
        
        # Check for empty inputs
        if y_true.size == 0 or y_prob.size == 0:
            return None
        
        # Handle binary case
        if y_prob.ndim == 1:
            y_prob = np.vstack([1 - y_prob, y_prob]).T
            
        # Dimension validation
        if len(y_true) != len(y_prob):
            raise ValueError("Number of samples in y_true and y_prob must match")
         
        # For multiclass case validate class labels
        if y_prob.shape[1] > 2:  # More than binary classification
            unique_labels = np.unique(y_true)
            if not (np.all(np.isin(unique_labels, np.arange(y_prob.shape[1])))):
                raise ValueError("Class labels in y_true must be integers in the range [0, n_classes-1].")
              
        # Check for invalid number of bins
        if n_bins <= 0:
            raise ValueError("Number of bins must be positive and higher than 0")
        
        n_classes = y_prob.shape[1]
        
        # Initialize arrays to store results
        mean_predicted_probs = np.zeros((n_bins, n_classes))
        true_probs = np.zeros((n_bins, n_classes))
        bin_counts = np.zeros((n_bins, n_classes))
        
        # Create bins
        bins = np.linspace(0, 1, n_bins + 1)
        
        # Calculate calibration curve for each class
        for class_idx in range(n_classes):
            class_confidences = y_prob[:, class_idx]
            class_true_labels = (y_true == class_idx).astype(int)  # True label for the current class
            bin_indices = np.digitize(class_confidences, bins) - 1
            
            for bin_idx in range(n_bins):
                mask = bin_indices == bin_idx
                # Handle empty bins 
                if np.sum(mask) > 0:
                    bin_counts[bin_idx, class_idx] = np.sum(mask)
                    mean_predicted_probs[bin_idx, class_idx] = np.mean(class_confidences[mask])
                    true_probs[bin_idx, class_idx] = np.mean(class_true_labels[mask])
                else:
                    # Fill empty bins with 0 instead of NaN
                    mean_predicted_probs[bin_idx, class_idx] = 0.0
                    true_probs[bin_idx, class_idx] = 0.0       
            # Log the results for the current class
            logger.info(f"\nCalibration Results for Class {class_idx} ({n_bins} bins):")
            for i in range(n_bins):
                if not np.isnan(mean_predicted_probs[i, class_idx]):
                    logger.info(f"Bin {i+1}: pred={mean_predicted_probs[i, class_idx]:.3f}, true={true_probs[i, class_idx]:.3f}, n={int(bin_counts[i, class_idx])}")
        
        return mean_predicted_probs, true_probs, bin_counts
        
    except Exception as e:
        logger.error(f"Error in multiclass calibration curve calculation: {str(e)}")
        raise


from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier, AdaBoostClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import SVC
from sklearn.naive_bayes import GaussianNB
from xgboost import XGBClassifier
from enum import Enum


class ProbabilisticSVC(SVC):
    """Wrapper for SVC that ensures probability estimation is enabled"""
    def __init__(self):
        super().__init__(probability=True)

class TabularModelsEnum(Enum):
    """
    Enum class for representing different machine learning model architectures for tabular data.
    Provides an interface to easily select and initialize popular classifiers.
    """
    RANDOM_FOREST = RandomForestClassifier  # Random Forest Classifier
    XGBOOST = XGBClassifier  # XGBoost Classifier
    SVM = ProbabilisticSVC  # Support Vector Machine Classifier
    NAIVE_BAYES = GaussianNB  # Gaussian Naive Bayes Classifier
    DECISION_TREE = DecisionTreeClassifier  # Decision Tree Classifier
    GRADIENT_BOOSTING = GradientBoostingClassifier  # Gradient Boosting Classifier
    ADABOOST = AdaBoostClassifier  # AdaBoost Classifier

    def __call__(self, *args, **kwargs):
        """
        Make the enum callable to directly call the model function.

        Args:
            *args: Variable positional arguments to pass to the model initialization.
            **kwargs: Variable keyword arguments to pass to the model initialization.

        Returns:
            An instance of the selected model initialized with the provided arguments.
        """
        return self.value(*args, **kwargs)  # Initialize and return the model instance
    
    
    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"




import random
import numpy as np
import torch
import logging
from abc import ABC, abstractmethod

from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager


config_manager = ConfigurationManager()
logging.basicConfig(level=logging.INFO, force=config_manager.logging)

class CalibratorBase(ABC):
    """
    Abstract base class for calibration methods.
    Provides a standardized interface for calibration models.
    """
    
    # Load default configurations
    DEFAULT_SEED = config_manager.random_seed  # Default random seed
    
    def __init__(self, seed=DEFAULT_SEED):
        self.fitted = False
        self.seed = seed
        self.set_random_seed(seed)
        
        # Initialize logger
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logs = True
        
        # Initialize metadata dictionary
        self.metadata = {
            "calibration_type": self.__class__.__name__.lower().replace('calibrator', ''),
            "params": {},
            "dataset_info": {}
        }
    
    def set_random_seed(self, seed: int):
        """
        Sets a fixed random seed for reproducibility.

        Args:
            seed (int): The random seed value.
        """
        if not isinstance(seed, int):
            raise TypeError(f"Seed must be an integer, got {type(seed)}.")

        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)
    
    @abstractmethod
    def fit(self, predictions: np.ndarray, ground_truth: np.ndarray):
        """
        Fits the calibration model.
        
        Args:
            predictions (np.ndarray): Model output probabilities or logits of shape (n_samples, n_classes).
            ground_truth (np.ndarray): True labels of shape (n_samples,).
        """
        pass
    
    @abstractmethod
    def predict(self, test_data: np.ndarray) -> np.ndarray:
        """
        Generates calibrated probabilities.
        
        Args:
            test_data (np.ndarray): Uncalibrated model outputs of shape (n_samples, n_classes).
        
        Returns:
            np.ndarray: Calibrated probabilities of shape (n_samples, n_classes).
        """
        pass
    
    def validate_inputs(self, predictions, ground_truth=None):
        """
        Validates input shapes and types.
        
        Args:
            predictions (np.ndarray): Model output logits or probabilities.
            ground_truth (np.ndarray): True labels.
        
        Returns:
            tuple: Validated predictions and ground truth labels.
        
        Raises:
            ValueError: If inputs do not match expected formats.
        """
        predictions = np.asarray(predictions)
        ground_truth = np.asarray(ground_truth)
        
        if predictions.ndim != 2 or ground_truth.ndim != 1:
            raise ValueError("Invalid input shapes: Predictions must be (n_samples, n_classes), labels must be (n_samples,).")
        
        if not np.all(np.isfinite(predictions)):
            raise ValueError("Predictions must be finite real numbers.")
        
        if ground_truth.dtype.kind not in {'i', 'u'}:
            raise ValueError("Ground truth labels must be integers.")
        
        if predictions.shape[0] != ground_truth.shape[0]:
            raise ValueError("Mismatched dimensions: predictions and labels must have the same number of samples.")
        
        return predictions, ground_truth
    


from itertools import product

from smartcal.config.enums.calibration_algorithms_enum import CalibrationAlgorithmTypesEnum
from smartcal.config.enums.calibration_hyperparameters import CalibrationHyperparameters


def get_calibration_hyperparameters(cal_algo: CalibrationAlgorithmTypesEnum):
    """
    Retrieves the hyperparameters configuration for a given calibration algorithm.
    
    Maps each calibration algorithm to its specific hyperparameters using the
    CalibrationHyperparameters enum values.
    
    Args:
        cal_algo (CalibrationAlgorithmTypesEnum): The calibration algorithm
        
    Returns:
        dict: Hyperparameters configuration for the algorithm
                Empty dict if algorithm has no hyperparameters
    """

    hyperparameters = {
            CalibrationAlgorithmTypesEnum.BETA: {
                'model_type': CalibrationHyperparameters.model_type.value
            },
            CalibrationAlgorithmTypesEnum.EMPIRICALBINNING: {
                'n_bins': CalibrationHyperparameters.n_bins.value
            },
            CalibrationAlgorithmTypesEnum.DIRICHLET: {
                'lr': CalibrationHyperparameters.lr.value,
                'max_iter': CalibrationHyperparameters.max_itr.value
            },
            CalibrationAlgorithmTypesEnum.MATRIXSCALING: {
                'lr': CalibrationHyperparameters.lr.value,
                'max_iter': CalibrationHyperparameters.max_itr.value
            },
            CalibrationAlgorithmTypesEnum.META: {
                'calibrator_type': CalibrationHyperparameters.calibrator_type_meta.value,
                'alpha': CalibrationHyperparameters.alpha.value,
                'acc': CalibrationHyperparameters.acc.value
            },
            CalibrationAlgorithmTypesEnum.TEMPERATURESCALING: {
                'initial_T': CalibrationHyperparameters.initial_T.value,
                'lr_tempscaling': CalibrationHyperparameters.lr_tempscaling.value,
                'max_iter_tempscaling': CalibrationHyperparameters.max_iter_tempscaling.value
            },
            CalibrationAlgorithmTypesEnum.VECTORSCALING: {
                'lr': CalibrationHyperparameters.lr.value,
                'max_iter': CalibrationHyperparameters.max_itr.value
            },
            CalibrationAlgorithmTypesEnum.ISOTONIC: {},
   
            CalibrationAlgorithmTypesEnum.PLATT: {
                'calibrator_type': CalibrationHyperparameters.calibrator_type_platt.value,
                'num_bins': CalibrationHyperparameters.num_bins.value
            },
            CalibrationAlgorithmTypesEnum.HISTOGRM: {
                'calibrator_type': CalibrationHyperparameters.calibrator_type_histogram.value,
                'num_bins': CalibrationHyperparameters.num_bins.value
            },
            CalibrationAlgorithmTypesEnum.AdaptiveTemperatureScaling: {
                'mode': CalibrationHyperparameters.adapt_temp_scaling_mode.value,
                'confidence_bins': CalibrationHyperparameters.adapt_temp_scaling_bins.value,
                'entropy_bins': CalibrationHyperparameters.adapt_temp_scaling_bins.value,
                'initial_T': CalibrationHyperparameters.initial_T.value,
                'lr_tempscaling': CalibrationHyperparameters.lr.value
            },
            CalibrationAlgorithmTypesEnum.MIXANDMATCH: {
                'parametric_calibrator': CalibrationHyperparameters.mix_match_parametric.value,
                'nonparametric_calibrator': CalibrationHyperparameters.mix_match_nonparametric.value
            }
        }
    return hyperparameters.get(cal_algo, {})

def make_hashable(item):
    """Recursively convert nested dictionaries to frozensets."""
    if isinstance(item, dict):
        return frozenset((k, make_hashable(v)) for k, v in item.items())
    elif isinstance(item, list):
        return tuple(make_hashable(v) for v in item)
    return item

def get_all_calibrator_combinations():
    """Generate all possible calibrator and hyperparameter combinations using function calls."""
    calibration_algorithms = list(CalibrationAlgorithmTypesEnum)
    calibrator_hyperparameters = []

    for cal in calibration_algorithms:
        calibrator_hyperparameters.append(get_calibration_hyperparameters(cal))

    combinations_dicts = []
    for algo, params in zip(calibration_algorithms, calibrator_hyperparameters):
        if algo.name == "META":
            if "alpha" in params and "acc" in params:
                for alpha_value in params["alpha"]:
                    combinations_dicts.append({
                        "Calibration_Algorithm": algo.name, 
                        "alpha": alpha_value, 
                        "acc": None
                    })
                for acc_value in params["acc"]:
                    combinations_dicts.append({
                        "Calibration_Algorithm": algo.name, 
                        "alpha": None, 
                        "acc": acc_value
                    })
        elif algo.name == 'AdaptiveTemperatureScaling':
            if params.get('mode') == 'entropy':
                params['confidence_bins'] = None
            elif params.get('mode') == 'linear':
                params['entropy_bins'] = None
            keys, values = zip(*params.items())
            for combination in product(*values):
                combinations_dicts.append({
                    "Calibration_Algorithm": algo.name, 
                    **dict(zip(keys, combination))
                })
        elif params:
            keys, values = zip(*params.items())
            for combination in product(*values):
                combinations_dicts.append({
                    "Calibration_Algorithm": algo.name, 
                    **dict(zip(keys, combination))
                })
        else:
            combinations_dicts.append({"Calibration_Algorithm": algo.name})

    # Ensure combinations are unique
    unique_combinations = set(make_hashable(comb) for comb in combinations_dicts)
    combinations_dicts = [dict(comb) for comb in unique_combinations]

    return combinations_dicts


from transformers import AutoModelForSequenceClassification
import fasttext
from enum import Enum


class ModelType(Enum):
    BERT = "BERT"
    WordEmbeddingModel = "WordEmbeddingModel"


class LanguageModelsEnum(Enum):
    # MOBILEBERT = ("google/mobilebert-uncased", AutoModelForSequenceClassification, ModelType.BERT)
    TINYBERT = ("huawei-noah/TinyBERT_General_4L_312D", AutoModelForSequenceClassification, ModelType.BERT)
    FASTTEXT = ("fasttext", None, ModelType.WordEmbeddingModel)  # No pretrained model class for FastText

    def __init__(self, model_name, model_class, model_type):
        self._model_name = model_name
        self._model_class = model_class
        self._model_type = model_type

    @property
    def model_type(self):
        """Return the broader model family/type (e.g., BERT or WordEmbeddingModel)."""
        return self._model_type

    @property
    def model_name(self):
        """Return the pretrained model identifier or 'fasttext' for FastText models."""
        return self._model_name

    def __call__(self, *args, **kwargs):
        """
        Make the enum callable to initialize the corresponding model.

        Returns:
            - A pretrained Hugging Face model for BERT-based models.
            - A placeholder FastText model for fasttext.
        """
        if self._model_class:
            return self._model_class.from_pretrained(self._model_name, *args, **kwargs)
        else:
            return fasttext.FastText._FastText()  # Placeholder FastText model

    def __str__(self):
        """Returns the string representation of the model name."""
        return self._model_name
    
    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"




import numpy as np
import logging


logger = logging.getLogger(__name__)

def calculate_brier_score(y_true, y_prob):
    """
    Calculate Multiclass Brier Score for classification predictions.
    
    Parameters:
    -----------
    y_true : array-like
        True labels (can be binary or multiclass)
    y_prob : array-like
        Predicted probabilities for all classes
        Shape should be (n_samples, n_classes)
        
    Returns:
    --------
    float or None
        Multiclass Brier score (lower is better, ranges from 0 to 1)
        Returns None if inputs are invalid.
    """
    try:
        y_true = np.array(y_true).astype(int)
        y_prob = np.array(y_prob)
        
        # Check for empty inputs
        if y_true.size == 0 or y_prob.size == 0:
           raise ValueError("Input arrays cannot be empty")
        
        # Handle binary case
        if y_prob.ndim == 1:
            y_prob = np.vstack([1 - y_prob, y_prob]).T
            
        n_samples, n_classes = y_prob.shape
        
        # Strict dimension validation
        if len(y_true) != n_samples:
            raise ValueError("Number of samples in y_true and y_prob must match")
            
        # For multiclass case, validate class labels
        if y_prob.shape[1] > 2:  # More than binary classification
            unique_labels = np.unique(y_true)
            if not (np.all(np.isin(unique_labels, np.arange(y_prob.shape[1])))):
                raise ValueError("Class labels in y_true must be integers in the range [0, n_classes-1].")
                
        # Convert y_true to one-hot encoding
        y_true_one_hot = np.zeros((n_samples, n_classes))
        y_true_one_hot[np.arange(n_samples), y_true] = 1
        
         # Calculate multiclass Brier score
        brier_score = np.mean(np.sum((y_prob - y_true_one_hot) ** 2, axis=1))
        
        logger.info(f"Brier Score: {brier_score:.4f}")

        return brier_score
        
    except Exception as e:
        logger.error(f"Error in multiclass Brier score calculation: {str(e)}")
        raise
    


import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.utils.timer import time_operation


config_manager = ConfigurationManager()

class DirichletCalibrator(CalibratorBase):
    """
    Implements Dirichlet Calibration for multi-class calibration.
    Adjusts model confidence using a full-rank transformation matrix and bias term.
    """
    def __init__(self, lr=config_manager.lr_tempscaling, max_iter=config_manager.max_iter_tempscaling, seed=CalibratorBase.DEFAULT_SEED):
        super().__init__(seed)
        self.lr = lr
        self.max_iter = max_iter
        self.transformation_matrix = None
        self.bias = None
        self.timing = {}
    
    @time_operation
    def fit(self, predictions, ground_truth):
        """
        Trains the Dirichlet calibration model.
        
        Args:
            predictions (np.ndarray): Model logits of shape (n_samples, n_classes).
            ground_truth (np.ndarray): True labels of shape (n_samples,).
        """
        self.set_random_seed(self.seed)
        predictions, ground_truth = self.validate_inputs(predictions, ground_truth)
        predictions = torch.tensor(predictions, dtype=torch.float32)
        ground_truth = torch.tensor(ground_truth, dtype=torch.long)

        num_classes = predictions.shape[1]
        self.transformation_matrix = nn.Parameter(torch.eye(num_classes, dtype=torch.float32, requires_grad=True))
        self.bias = nn.Parameter(torch.zeros(num_classes, dtype=torch.float32, requires_grad=True))
        
        optimizer = optim.Adam([self.transformation_matrix, self.bias], lr=self.lr)
        criterion = nn.CrossEntropyLoss()
        
        for _ in range(self.max_iter):
            optimizer.zero_grad()
            transformed_logits = torch.matmul(predictions, self.transformation_matrix) + self.bias
            loss = criterion(torch.nn.functional.log_softmax(transformed_logits, dim=1), ground_truth)
            loss.backward()
            optimizer.step()
        
        self.fitted = True
        
        self.metadata["dataset_info"].update({
            "n_samples": predictions.shape[0],
            "n_classes": num_classes,
        })
        self.metadata["params"].update({
            "learning_rate": self.lr,
            "max_iterations": self.max_iter,
        })
    
    @time_operation
    def predict(self, test_data):
        """
        Applies Dirichlet calibration to logits.

        Args:
            test_data (np.ndarray): Logits of shape (n_samples, n_classes).

        Returns:
            np.ndarray: Calibrated probabilities of shape (n_samples, n_classes).
        """
        if not self.fitted:
            raise RuntimeError("Model must be fitted before prediction.")

        if not isinstance(test_data, np.ndarray):
            raise ValueError("Test data must be a NumPy array.")

        if test_data.ndim != 2:
            raise ValueError(f"Test data must be a 2D NumPy array, but got shape {test_data.shape}.")

        test_data = torch.tensor(test_data, dtype=torch.float32)

        # **NEW: Shape validation**
        if test_data.shape[1] != self.transformation_matrix.shape[0]:
            raise ValueError(
                f"Mismatch in number of classes: expected {self.transformation_matrix.shape[0]}, but got {test_data.shape[1]}"
            )

        calibrated_logits = torch.matmul(test_data, self.transformation_matrix) + self.bias
        return torch.nn.functional.softmax(calibrated_logits, dim=1).detach().numpy()
    
    def get_timing(self):
        """Return the timing information for fit and predict operations."""
        return self.timing



import numpy as np


def convert_one_hot_to_labels(y_true: np.ndarray) -> np.ndarray:
    """
    Convert one-hot encoded labels to single-label format.

    :param y_true: A NumPy array containing ground truth labels (one-hot or single-label).
    :return: A NumPy array with single-label format.
    """
    if y_true.ndim == 2 and y_true.shape[1] > 1:  # Detect one-hot encoding
        return np.argmax(y_true, axis=1)  # Convert to single-label format
    return y_true  # Return unchanged if already single-label

def normalize_probabilities(probs, tolerance=0):
    """
    Normalize a 1D or 2D array of prediction probabilities so that each row sums to 1.
    If the input is already normalized (within a small tolerance), it is returned as is.

    Args:
        probs (list or np.array): A 1D or 2D array of prediction probabilities.
        tolerance (float): Allowed deviation from 1.0 to consider the input already normalized.
        
    Returns:
        np.array: Normalized probabilities if needed, else original input.
    """
    probs = np.array(probs, dtype=np.float64)  # Convert to NumPy array
    
    if np.any(probs < 0):  # Ensure no negative values
        raise ValueError("Probabilities must be non-negative.")
    
    total = np.sum(probs, axis=-1, keepdims=True)  # Sum along the last axis
    
    # Check if already normalized within tolerance
    if np.all(np.abs(total - 1) < tolerance):
        return probs  # Return as is if already normalized
    
    if np.any(total == 0):  # Prevent division by zero
        raise ValueError("Sum of probabilities cannot be zero for any row.")
    
    return probs / total  # Normalize only if needed


from sklearn.isotonic import IsotonicRegression
import numpy as np

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.utils.timer import time_operation


class IsotonicCalibrator(CalibratorBase):
    """
    Isotonic calibration for both binary and multi-class classification.
    Binary: Only calibrates the probability for the positive class.
    Multi-class: One-vs-rest approach, then re-normalizes probabilities.
    """

    def __init__(self, seed=CalibratorBase.DEFAULT_SEED):
        super().__init__(seed=seed)
        self.models = {}        # dict: class_idx -> IsotonicRegression instance
        self.n_classes_ = None
        self.timing = {} 
        self.metadata.update({
            "calibration_type": "isotonic",
        })

    @time_operation
    def fit(self, predictions: np.ndarray, ground_truth: np.ndarray):
        """
        predictions : shape (n_samples, n_classes)
            Assumed to be probabilities, not raw logits.

        ground_truth : shape (n_samples,)
            Class labels.
        """
        predictions, ground_truth = super().validate_inputs(predictions, ground_truth)

        self.n_classes_ = predictions.shape[1]
        n_samples = predictions.shape[0]
        self.metadata["dataset_info"].update({
            "n_samples": n_samples,
            "n_classes": self.n_classes_
        })
        
        self.logger.info(f"IsotonicCalibration.fit with {n_samples} samples, {self.n_classes_} classes")

        # For binary classification: we only calibrate the 'positive' column
        # For multi-class: do one-vs-rest and re-normalize
        if self.n_classes_ == 2:
            # We consider class '1' as "positive"
            # Extract the probability for the positive class
            p_pos = predictions[:, 1]
            # Convert ground_truth to a binary [0,1] vector
            bin_labels = (ground_truth == 1).astype(float)

            # Create and fit an isotonic regression model
            iso_model = IsotonicRegression(out_of_bounds='clip')
            iso_model.fit(p_pos, bin_labels)
            self.models[1] = iso_model

        else:
            # Multi-class: one vs rest for each class
            for class_idx in range(self.n_classes_):
                bin_labels = (ground_truth == class_idx).astype(float)
                # Probability for "this" class
                p_class = predictions[:, class_idx]

                iso_model = IsotonicRegression(out_of_bounds='clip')
                iso_model.fit(p_class, bin_labels)
                self.models[class_idx] = iso_model

        self.fitted = True
        self.metadata["params"]["n_models"] = len(self.models)
        self.logger.info(f"IsotonicCalibration fitted {len(self.models)} model(s)")

    @time_operation
    def predict(self, test_data: np.ndarray) -> np.ndarray:
        """
        test_data : shape (n_samples, n_classes)
            Probabilities to be calibrated (not raw logits).
        Returns
        -------
        calibrated_probs : shape (n_samples, n_classes)
        """
        if not self.fitted:
            raise RuntimeError("Calibrator not fitted. Call `fit` first.")

        if test_data.shape[1] != self.n_classes_:
            raise ValueError("Number of columns in `test_data` must match training.")

        n_samples = test_data.shape[0]
        calibrated_probs = np.zeros_like(test_data)

        if self.n_classes_ == 2:
            # For binary, we only have an Isotonic model for the 'positive' class
            iso_model = self.models[1]
            p_pos = test_data[:, 1]
            calibrated_pos = iso_model.predict(p_pos)
            calibrated_neg = 1.0 - calibrated_pos

            calibrated_probs[:, 0] = calibrated_neg
            calibrated_probs[:, 1] = calibrated_pos

        else:
            # Multi-class: apply each isotonic model, then re-normalize
            for class_idx in range(self.n_classes_):
                iso_model = self.models[class_idx]
                p_class = test_data[:, class_idx]
                calibrated_probs[:, class_idx] = iso_model.predict(p_class)

            # renormalize row-wise so they sum to 1
            row_sums = calibrated_probs.sum(axis=1, keepdims=True)
            # avoid division by zero
            np.clip(row_sums, 1e-15, None, out=row_sums)
            calibrated_probs /= row_sums

        return calibrated_probs
    
    def get_timing(self):
        """Return the timing information for fit and predict operations."""
        return self.timing



from typing import Dict, Callable, Union, List, Any
import numpy as np
import logging
from bayes_opt import BayesianOptimization
from sklearn.metrics import f1_score, log_loss
from sklearn.model_selection import StratifiedKFold
import time

from smartcal.config.enums.calibration_algorithms_enum import CalibrationAlgorithmTypesEnum
from smartcal.utils.cal_metrics import compute_calibration_metrics
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.bayesian_optimization.bayesian_hyperparameters import param_mappings, param_spaces
from smartcal.meta_features_extraction.meta_features_extraction import MetaFeaturesExtractor

config_manager = ConfigurationManager()


def get_metric_value(metrics: Dict[str, Any], y_true: np.array, y_pred: np.array, metric_name: str) -> float:
    """
    Get a specific metric value from computed metrics or calculate it if needed.
    
    Args:
        metrics: Dictionary of pre-computed calibration metrics
        y_true: True labels
        y_pred: Predicted probabilities
        metric_name: Name of the metric to retrieve
        
    Returns:
        float: The value of the requested metric
    """
    metric_evaluators = {
        'ECE': lambda m, y_t, y_p: m['ece'],
        'MCE': lambda m, y_t, y_p: m['mce'],
        'brier_score': lambda m, y_t, y_p: m['brier_score'],
        'ConfECE': lambda m, y_t, y_p: np.mean(m['conf_ece']),
        'log_loss': lambda m, y_t, y_p: log_loss(y_t, y_p)
    }
    
    if metric_name not in metric_evaluators:
        raise ValueError(f"Unsupported metric: {metric_name}")
    
    return metric_evaluators[metric_name](metrics, y_true, y_pred)


class CalibrationOptimizer:
    def __init__(self, meta_model, k_folds=config_manager.cal_tune_kfolds):
        """
        Initialize the CalibrationOptimizer with a meta-model and cross-validation settings.

        Args:
            meta_model: A model capable of recommending calibration algorithms
            k_folds (int, optional): Number of cross-validation folds.
        """
        self.meta_model = meta_model
        self.k_folds = k_folds

    def get_calibrator_recommendations(self, y_true, y_pred):
        """
        Gets the calibrator recommendations based on predicted probabilities and true labels.

        This function extracts meta-features from the given calibration data and then
        uses the meta-model to predict which calibrators should be used, and their associated confidence scores.

        Args:
            y_true (np.array): True labels of the calibration dataset.
            y_pred (np.array): Predicted probabilities from a classifier.

        Returns:
            list: A list of tuples, each containing a calibrator name and its corresponding normalized confidence score.
        """
        extractor = MetaFeaturesExtractor()
        meta_features = extractor.process_features(y_true, y_pred)
        recommendations = self.meta_model.predict_best_model(meta_features)
        
        # Normalize confidence scores to sum to 1
        calibrators, scores = zip(*recommendations)
        normalized_scores = np.array(scores) / sum(scores)
        
        logging.info(f"Recommendations: {recommendations}")
        logging.info(f"Normalized Scores: {normalized_scores}")
        
        return list(zip(calibrators, normalized_scores))

    def allocate_iterations(self, normalized_scores, total_iterations):
        """
        Allocates the total number of iterations to each calibrator based on its normalized confidence score.

        This ensures that the sum of all allocations equals the total_iterations and
        tries to allocate iterations proportional to confidence.

        Args:
            normalized_scores (list): List of normalized confidence scores for each calibrator.
            total_iterations (int): Total number of iterations to allocate.

        Returns:
            np.array: The final allocation of iterations for each calibrator.
        """

        logging.info(f"Normalized scores: {normalized_scores}")

        # Ensure normalized_scores is a numpy array
        normalized_scores = np.array(normalized_scores)

        # Calculate raw allocations
        raw_allocations = normalized_scores * total_iterations
        logging.info(f"Raw allocations: {raw_allocations}")

        # Floor the raw allocations
        allocations = np.floor(raw_allocations).astype(int)
        logging.info(f"After floor: {allocations}")

        # Calculate remainders
        remainders = raw_allocations - allocations
        logging.info(f"Remainders: {remainders}")

        # Calculate remaining iterations
        remaining_iterations = total_iterations - np.sum(allocations)
        logging.info(f"Remaining iterations: {remaining_iterations}")

        # Allocate remaining iterations to the calibrators with the largest remainders
        indices = np.argsort(-remainders)[:remaining_iterations]
        logging.info(f"Indices for remaining: {indices}")

        allocations[indices] += 1
        logging.info(f"Final allocations: {allocations}")

        return allocations

    def optimize_calibrator(self, calibrator_name: str, X_cal, y_cal, n_iterations: int, metric: str = config_manager.metric) -> Dict:
        """
        Optimize hyperparameters for a specific calibration algorithm using Bayesian Optimization.

        This method handles mixed parameter types (continuous, categorical, integer)
        and uses cross-validation to evaluate calibrator performance.

        Args:
            calibrator_name (str): Name of the calibration algorithm
            X_cal (array-like): Calibration set features
            y_cal (array-like): Calibration set labels
            n_iterations (int): Number of Bayesian optimization iterations

        Returns:
            Dict: Optimization results including best parameters, score, and metrics
        """

        calibrator_class = CalibrationAlgorithmTypesEnum[calibrator_name]

        bayesian_opt_start_time = time.perf_counter()

        def objective(**params):
            # Convert parameters back to their original types
            converted_params = {}
            mapping = param_mappings.get(calibrator_name, {})

            for param_name, param_value in params.items():
                if param_name in mapping:
                    # Categorical parameter - map back to string
                    choices = list(mapping[param_name].keys())
                    idx = int(round(param_value))
                    converted_params[param_name] = choices[idx % len(choices)]
                elif param_name.endswith('_iters') or param_name.endswith('_iter') or 'max_iter' in param_name:
                    # Integer parameter
                    converted_params[param_name] = int(round(param_value))
                elif param_name == 'n_bins' or param_name == 'num_bins' or 'bins' in param_name:
                    # Integer parameter for bins
                    converted_params[param_name] = int(round(param_value))
                else:
                    # Continuous parameter
                    converted_params[param_name] = param_value

            calibrator = calibrator_class(**converted_params)
            scores = []

            kf = StratifiedKFold(n_splits=self.k_folds, shuffle=True, random_state=config_manager.random_seed)
            for train_idx, val_idx in kf.split(X_cal, y_cal):
                train_idx = np.array(train_idx, dtype=int)
                val_idx = np.array(val_idx, dtype=int)

                X_train, X_val = X_cal[train_idx], X_cal[val_idx]
                y_train, y_val = y_cal[train_idx], y_cal[val_idx]

                calibrator.fit(X_train, y_train)
                calibrated_probs = calibrator.predict(X_val)
                calibrated_predicted_label = np.argmax(calibrated_probs, axis=1).tolist()
                metrics = compute_calibration_metrics(calibrated_probs, calibrated_predicted_label, y_val)
                
                # Get the metric value using the centralized function
                metric_value = get_metric_value(metrics, y_val, calibrated_probs, metric)
                scores.append(metric_value)

            return -np.mean(scores)

        # Get the parameter space for this calibrator
        pbounds = param_spaces.get(calibrator_name, {})

        if not pbounds:
            # No parameters to optimize
            calibrator = calibrator_class()
            calibrator.fit(X_cal, y_cal)
            calibrated_probs = calibrator.predict(X_cal)
            calibrated_predicted_label = np.argmax(calibrated_probs, axis=1).tolist()
            metrics = compute_calibration_metrics(calibrated_probs, calibrated_predicted_label, y_cal)

            bayesian_opt_end_time = time.perf_counter()
            bayesian_opt_time = bayesian_opt_end_time - bayesian_opt_start_time

            return {
                'best_params': {},
                'best_score': metrics['ece'],
                'full_metrics': metrics,
                'bayesian_optimization_time': bayesian_opt_time
            }

        optimizer = BayesianOptimization(
            f=objective,
            pbounds=pbounds,
            random_state=config_manager.random_seed
        )

        init_points = min(3, n_iterations)
        n_iter = max(0, n_iterations - init_points)
        optimizer.maximize(init_points=init_points, n_iter=n_iter)

        bayesian_opt_end_time = time.perf_counter()
        bayesian_opt_time = bayesian_opt_end_time - bayesian_opt_start_time

        # Convert best parameters back to original types
        best_params = {}
        mapping = param_mappings.get(calibrator_name, {})

        for param_name, param_value in optimizer.max['params'].items():
            if param_name in mapping:
                choices = list(mapping[param_name].keys())
                idx = int(round(param_value))
                best_params[param_name] = choices[idx % len(choices)]
            elif param_name.endswith('_iters') or param_name.endswith('_iter') or 'max_iter' in param_name:
                best_params[param_name] = int(round(param_value))
            elif param_name == 'n_bins' or param_name == 'num_bins' or 'bins' in param_name:
                best_params[param_name] = int(round(param_value))
            else:
                best_params[param_name] = param_value

        # Fit with the best parameters to get full metrics
        best_calibrator = calibrator_class(**best_params)
        best_calibrator.fit(X_cal, y_cal)
        calibrated_probs = best_calibrator.predict(X_cal)
        calibrated_predicted_label = np.argmax(calibrated_probs, axis=1).tolist()
        full_metrics = compute_calibration_metrics(calibrated_probs, calibrated_predicted_label, y_cal)

        return {
            'best_params': best_params,
            'best_score': -optimizer.max['target'],
            'full_metrics': full_metrics,
            'bayesian_optimization_time': bayesian_opt_time
        }

    def run_optimization(self, dataset_results: Dict, total_iterations: int = 30, metric: str = config_manager.metric):
        """Run the complete optimization process, saving only the best calibrator's results"""
        results = {}

        try:
            # Ensure X_cal and y_cal are NumPy arrays
            X_cal = np.array(dataset_results['calibration_set_predicted_probabilities'])
            y_cal = np.array(dataset_results['calibration_set_true_labels'])

            X_test = np.array(dataset_results['test_set_predicted_probabilities'])
            y_test = np.array(dataset_results['test_set_true_labels'])

            # Validate shapes
            if X_cal.ndim != 2 or X_test.ndim != 2:
                raise ValueError("Input data must be 2D arrays (samples x features).")

            # Get recommendations and allocate iterations
            recommendations = self.get_calibrator_recommendations(y_cal, X_cal)
            normalized_scores = [score for _, score in recommendations]
            iteration_allocations = self.allocate_iterations(normalized_scores, total_iterations)

            best_calibrator = None
            best_metric_value = float('inf')
            best_hyperparameters = "None"

            # Store the metric being optimized
            results["optimization_metric"] = metric

            # Optimize each recommended calibrator
            for (calibrator_name, confidence), n_iterations in zip(recommendations, iteration_allocations):
                if n_iterations > 0:
                    try:
                        optimization_result = self.optimize_calibrator(
                            calibrator_name, X_cal, y_cal, n_iterations, metric
                        )

                        # Skip if optimization failed
                        if optimization_result is None or optimization_result.get('full_metrics') is None:
                            continue

                        # Instantiate and fit the calibrator with best parameters
                        current_params = optimization_result['best_params'].copy()

                        # For params saving purposes
                        param_str = ", ".join(
                            [f"{k}={v}" for k, v in current_params.items()]) if current_params else "None"

                        try:
                            calibrator = CalibrationAlgorithmTypesEnum[calibrator_name](**current_params)
                            calibrator.fit(X_cal, y_cal)

                            # Evaluate on calibration set
                            cal_probs = calibrator.predict(X_cal)
                            cal_predicted_label = np.argmax(cal_probs, axis=1).tolist()
                            cal_metrics = compute_calibration_metrics(cal_probs, cal_predicted_label, y_cal)
                            
                            # Get the specific metric using the centralized function
                            current_metric_value = get_metric_value(cal_metrics, y_cal, cal_probs, metric)

                            # For all our metrics, lower is better
                            if current_metric_value < best_metric_value:
                                best_metric_value = current_metric_value
                                best_calibrator = calibrator
                                best_hyperparameters = param_str

                                # Store metrics
                                results["cal_ece"] = float(cal_metrics['ece'])
                                results["cal_mce"] = float(cal_metrics['mce'])
                                results["cal_conf_ece"] = list(cal_metrics['conf_ece'])
                                results["cal_brier_score"] = float(cal_metrics['brier_score'])
                                results["cal_log_loss"] = float(log_loss(y_cal, cal_probs))
                                results["cal_f1_score_macro"] = float(
                                    f1_score(y_cal, cal_probs.argmax(axis=1), average='macro'))
                                results["cal_f1_score_micro"] = float(
                                    f1_score(y_cal, cal_probs.argmax(axis=1), average='micro'))
                                results["cal_calibration_curve_mean_predicted_probs"] = cal_metrics[
                                    'calibration_curve_mean_predicted_probs']
                                results["cal_calibration_curve_true_probs"] = cal_metrics[
                                    'calibration_curve_true_probs']
                                results["cal_calibration_curve_bin_counts"] = cal_metrics[
                                    'calibration_curve_bin_counts']

                                # Store the optimized metric value
                                results[f"best_{metric}_value"] = float(current_metric_value)

                                results["confidence_score"] = float(confidence)
                                results["allocated_iterations"] = int(n_iterations)
                                results["bayesian_optimization_time"] = optimization_result.get(
                                    'bayesian_optimization_time', None)

                        except Exception as e:
                            logging.warning(f"Failed to evaluate {calibrator_name}: {str(e)}")
                            continue

                    except Exception as e:
                        logging.warning(f"Optimization failed for {calibrator_name}: {str(e)}")
                        continue

            # Store best calibrator info
            results["best_calibrator"] = best_calibrator.__class__.__name__ if best_calibrator else "None"
            results["hyperparameters"] = best_hyperparameters

            # Evaluate best calibrator on test set if available
            if best_calibrator is not None:
                try:
                    test_probs = best_calibrator.predict(X_test)
                    test_predicted_label = np.argmax(test_probs, axis=1).tolist()
                    test_metrics_full = compute_calibration_metrics(test_probs, test_predicted_label, y_test)

                    # Store test metrics
                    results["test_ece"] = float(test_metrics_full['ece'])
                    results["test_mce"] = float(test_metrics_full['mce'])
                    results["test_conf_ece"] = list(test_metrics_full['conf_ece'])
                    results["test_brier_score"] = float(test_metrics_full['brier_score'])
                    results["test_log_loss"] = float(log_loss(y_test, test_probs))
                    results["test_f1_score_macro"] = float(f1_score(y_test, test_probs.argmax(axis=1), average='macro'))
                    results["test_f1_score_micro"] = float(f1_score(y_test, test_probs.argmax(axis=1), average='micro'))

                    # Store calibration curve data
                    results["test_calibration_curve_mean_predicted_probs"] = test_metrics_full[
                        'calibration_curve_mean_predicted_probs']
                    results["test_calibration_curve_true_probs"] = test_metrics_full['calibration_curve_true_probs']
                    results["test_calibration_curve_bin_counts"] = test_metrics_full['calibration_curve_bin_counts']

                except Exception as e:
                    logging.warning(f"Test evaluation failed: {str(e)}")
                    results["test_error"] = str(e)

        except Exception as e:
            logging.error(f"Critical error in run_optimization: {str(e)}")
            results["error"] = str(e)

        return results


import numpy as np
import bisect
from typing import List, TypeVar

from smartcal.calibration_algorithms.calibration_base import CalibratorBase
from smartcal.config.configuration_manager.configuration_manager import ConfigurationManager
from smartcal.utils.timer import time_operation


config_manager = ConfigurationManager()

class HistogramScalingCalibrator(CalibratorBase):
    """
    Implements histogram binning for multi-class probability calibration.
    """
    def __init__(self, num_bins=config_manager.n_bin_cal, seed=CalibratorBase.DEFAULT_SEED):
        """
        Initialize the Histogram Binning calibrator.
        :param num_bins: Number of bins to use for histogram binning.
        """
        super().__init__(seed)
        self.num_bins = num_bins
        self.histogram_models = None
        self.bins = None
        self.n_classes = None
        self.metadata.update({"calibration_type": "histogram_binning"})  # Store calibration type
        self.metadata["params"].update({"num_bins": self.num_bins})  
        self.timing = {}

    @time_operation
    def fit(self, logits, labels):
        """
        Train the Histogram Binning calibrator.
        :param logits: (n_samples, n_classes) array of model logits or confidence scores.
        :param labels: (n_samples,) array of true labels.
        """
        logits, labels = super().validate_inputs(logits, labels)  # Ensure input consistency
        logits = np.array(logits)
        labels = np.array(labels)
        n_samples, self.n_classes = logits.shape

        self.histogram_models = []
        self.bins = []

        # Log dataset info
        self.metadata["dataset_info"].update({"n_samples": n_samples, "n_classes": self.n_classes})

        # One-vs-All Histogram Binning
        for class_idx in range(self.n_classes):
            # Get class-wise probabilities
            class_probs = logits[:, class_idx]
            class_labels = (labels == class_idx).astype(float)  # Convert labels to binary (One-vs-All)

            # Compute bin edges
            try:
                bins = get_equal_bins(class_probs, num_bins=self.num_bins)
            except ValueError as e:
                self.logger.error(f"Bin computation failed for class {class_idx}: {e}")
                raise RuntimeError(f"Failed to compute bins for class {class_idx}. Reduce `num_bins`.")

            # Train histogram calibrator
            try:
                histogram_model = get_histogram_calibrator(class_probs, class_labels, bins)
            except ValueError as e:
                self.logger.error(f"Histogram calibration failed for class {class_idx}: {e}")
                raise RuntimeError(f"Histogram calibration failed for class {class_idx}.")

            self.bins.append(bins)
            self.histogram_models.append(histogram_model)

        self.fitted = True
        self.logger.info(f"HistogramCalibrator fitted with {self.num_bins} bins per class.")

    @time_operation
    def predict(self, logits):
        """
        Apply Histogram Binning calibration to logits.
        :param logits: (n_samples, n_classes) array of logits or confidence scores.
        :return: (n_samples, n_classes) array of calibrated probabilities.
        """
        if not self.fitted:
            raise RuntimeError("Calibrator not fitted. Call `fit()` first.")

        logits = np.array(logits)

        # Ensure input shape matches the number of classes
        if logits.shape[1] != self.n_classes:
            raise ValueError(f"Mismatch in number of classes: expected {self.n_classes}, got {logits.shape[1]}.")

        n_samples = logits.shape[0]
        calibrated_probs = np.zeros((n_samples, self.n_classes))

        # Apply histogram binning for each class
        for class_idx in range(self.n_classes):
            class_probs = logits[:, class_idx]
            calibrated_probs[:, class_idx] = self.histogram_models[class_idx](class_probs)

        # Normalize to ensure probabilities sum to 1
        row_sums = calibrated_probs.sum(axis=1, keepdims=True)
        calibrated_probs /= np.clip(row_sums, 1e-15, None)  # Avoid division by zero

        return calibrated_probs
    
    def get_timing(self):
        return self.timing

# Define data types
T = TypeVar('T')
eps = 1e-6

def split(sequence: List[T], parts: int) -> List[List[T]]:
    """
    Splits a sequence into approximately equal-sized parts.
    """
    assert parts <= len(sequence)
    array_splits = np.array_split(sequence, parts)
    return [list(l) for l in array_splits]

def get_bin(pred_prob: float, bins: List[float]) -> int:
    """
    Get the index of the bin that pred_prob belongs in.
    """
    assert 0.0 <= pred_prob <= 1.0
    assert bins[-1] == 1.0
    return bisect.bisect_left(bins, pred_prob)

def get_equal_bins(probs: List[float], num_bins: int = 10) -> List[float]:
    """
    Compute bin edges that contain approximately an equal number of samples.
    """
    sorted_probs = sorted(probs)
    if num_bins > len(sorted_probs):
        num_bins = len(sorted_probs)

    binned_data = split(sorted_probs, num_bins)
    bins = []
    for i in range(len(binned_data) - 1):
        last_prob = binned_data[i][-1]
        next_first_prob = binned_data[i + 1][0]
        bins.append((last_prob + next_first_prob) / 2.0)
    bins.append(1.0)
    return sorted(list(set(bins)))

def get_histogram_calibrator(model_probs, values, bins):
    """
    Creates a histogram-based calibration function.
    """
    binned_values = [[] for _ in range(len(bins))]

    for prob, value in zip(model_probs, values):
        bin_idx = get_bin(prob, bins)
        binned_values[bin_idx].append(float(value))

    def safe_mean(values, bin_idx):
        if len(values) == 0:
            return float(bins[bin_idx]) / 2.0 if bin_idx == 0 else float(bins[bin_idx] + bins[bin_idx - 1]) / 2.0
        return np.mean(values)

    bin_means = np.array([safe_mean(values, idx) for idx, values in enumerate(binned_values)])

    def calibrator(probs):
        indices = np.searchsorted(bins, probs)
        return bin_means[indices]

    return calibrator
